""" Contains various utility functions and utility classes. """

# clean
import datetime as dt
import gc
import hashlib
import inspect
import itertools
import json
import os
from dataclasses import dataclass
from functools import reduce as freduce
from functools import wraps
from timeit import default_timer as timer
from typing import Any, Dict, List, Optional, Tuple

import pandas as pd
import psutil
import pytz

from hisim import log
from hisim.simulationparameters import SimulationParameters

__authors__ = "Noah Pflugradt, Vitor Hugo Bellotto Zago"
__copyright__ = "Copyright 2021-2022, FZJ-IEK-3 "
__license__ = "MIT"
__version__ = "1"
__maintainer__ = "Noah Pflugradt"
__email__ = "n.pflugradt@fz-juelich.de"
__status__ = "development"


def get_input_directory() -> str:
    """Gets the absolute path to the inputs directory."""
    return os.path.join(hisim_abs_path, "inputs")


# Retrieves hisim directory absolute path
hisim_abs_path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))  # type: ignore
hisim_inputs = os.path.join(hisim_abs_path, "inputs")
hisim_results = os.path.join(hisim_abs_path, "results")
hisim_postprocessing_img = os.path.join(hisim_abs_path, "postprocessing", "report")  # noqa

HISIMPATH: Dict[str, Any] = {
    "inputs": hisim_inputs,
    "cache_dir": os.path.join(hisim_abs_path, "inputs", "cache"),
    "cache_indices": os.path.join(hisim_abs_path, "inputs", "cache", "cache_indices.json"),
    "cfg": os.path.join(hisim_abs_path, "inputs", "cfg.json"),
    "utsp_results": hisim_results,
    "utsp_example_results": os.path.join(hisim_inputs, "LPGResults_for_tests", "Results"),
    "utsp_example_reports": os.path.join(hisim_inputs, "LPGResults_for_tests", "Reports"),
    "housing": os.path.join(hisim_inputs, "housing", "data_processed", "episcope-tabula.csv"),
    "housing_reference_temperatures": os.path.join(
        hisim_inputs,
        "housing",
        "data_processed",
        "heating_reference_temperature_per_location.csv",
    ),
    "heater_efficiencies": os.path.join(
        hisim_inputs,
        "housing",
        "data_processed",
        "heater_efficiencies.csv",
    ),
    "fuel_costs": os.path.join(hisim_abs_path, "modular_household", "emission_factors_and_costs_fuels.csv"),
    "component_costs": os.path.join(hisim_abs_path, "modular_household", "emission_factors_and_costs_devices.csv"),
    "occupancy_scaling_factors_per_country": os.path.join(
        hisim_inputs, "loadprofiles", "WHY_reference_data", "scaling_factors_demand.csv"
    ),
    "occupancy": {
        "CHR01 Couple both at Work": {
            "number_of_residents": [
                os.path.join(
                    hisim_inputs,
                    "loadprofiles",
                    "predefined_lpg_household_chr01",
                    "data_processed",
                    "BodilyActivityLevel.High.HH1.json",
                ),
                os.path.join(
                    hisim_inputs,
                    "loadprofiles",
                    "predefined_lpg_household_chr01",
                    "data_processed",
                    "BodilyActivityLevel.Low.HH1.json",
                ),
            ],
            "electricity_consumption": os.path.join(
                hisim_inputs,
                "loadprofiles",
                "predefined_lpg_household_chr01",
                "data_processed",
                "SumProfiles.HH1.Electricity.csv",
            ),
            "electricity_consumption_without_washing_machine_and_dishwasher": os.path.join(
                hisim_inputs,
                "loadprofiles",
                "predefined_lpg_household_chr01",
                "data_processed",
                "SumProfiles.NoFlex.HH1.Electricity.csv",
            ),
            "heating_by_devices": os.path.join(
                hisim_inputs,
                "loadprofiles",
                "predefined_lpg_household_chr01",
                "data_processed",
                "SumProfiles.HH1.Inner Device Heat Gains.csv",
            ),
            "water_consumption": os.path.join(
                hisim_inputs,
                "loadprofiles",
                "predefined_lpg_household_chr01",
                "data_processed",
                "SumProfiles.HH1.Warm Water.csv",
            ),
        },
        "AVG": {
            "number_of_residents": [
                os.path.join(
                    hisim_inputs,
                    "loadprofiles",
                    "WHY_reference_data",
                    "BodilyActivityLevel.High.HH1.json",
                ),
                os.path.join(
                    hisim_inputs,
                    "loadprofiles",
                    "WHY_reference_data",
                    "BodilyActivityLevel.Low.HH1.json",
                ),
            ],
            "electricity_consumption": os.path.join(
                hisim_inputs,
                "loadprofiles",
                "WHY_reference_data",
                "AVG.csv",
            ),
            "water_consumption": os.path.join(
                hisim_inputs,
                "loadprofiles",
                "WHY_reference_data",
                "WarmWater.csv",
            ),
        },
    },
    "photovoltaic": {
        "sandia_modules_new": os.path.join(hisim_inputs, "photovoltaic", "data_processed", "sandia_modules_new.csv"),
        "sandia_modules": os.path.join(hisim_inputs, "photovoltaic", "data_processed", "sandia_modules.csv"),
        "sandia_inverters": os.path.join(hisim_inputs, "photovoltaic", "data_processed", "sandia_inverters.csv"),
        "cec_modules": os.path.join(hisim_inputs, "photovoltaic", "data_processed", "cec_modules.csv"),
        "cec_inverters": os.path.join(hisim_inputs, "photovoltaic", "data_processed", "cec_inverters.csv"),
    },
    "chp_system": os.path.join(hisim_inputs, "chp_system"),
    "smart_appliances": os.path.join(hisim_inputs, "smart_devices", "data_processed", "smart_devices.json"),
    "frank_data": os.path.join(
        hisim_inputs,
        "loadprofiles",
        "electrical-spaceheating-warmwater-photovoltaic_1-household",
        "data_raw",
        "VDI 4655",
    ),
    "report": os.path.join(hisim_abs_path, "results", "report.pdf"),
    "advanced_battery": {
        "parameter": os.path.join(hisim_abs_path, "inputs", "advanced_battery", "parameter", "PerModPAR.xlsx"),
        "reference_case": os.path.join(
            hisim_abs_path,
            "inputs",
            "advanced_battery",
            "reference_case",
            "ref_case_data.npz",
        ),
        "siemens_junelight": os.path.join(hisim_abs_path, "inputs", "advanced_battery", "Siemens_Junelight.npy"),
    },
    "LoadProfileGenerator_export_directory": os.path.join(os.path.join("D:", os.sep, "Work")),
    "bat_parameter": os.path.join(hisim_abs_path, "inputs", "advanced_battery", "Siemens_Junelight.npy"),
    "modular_household": os.path.join(hisim_abs_path, "modular_household"),
    "price_signal": {
        "PricePurchase": os.path.join(hisim_inputs, "price_signal", "PricePurchase.csv"),
        "FeedInTarrif": os.path.join(hisim_inputs, "price_signal", "FeedInTarrif.csv"),
    },
}


def load_smart_appliance(name):  # noqa
    """Loads file for a single smart appliance by name."""
    with open(HISIMPATH["smart_appliances"], encoding="utf-8") as filestream:
        data = json.load(filestream)
    return data[name]


def convert_lpg_timestep_to_utc(data: List[int], year: int, seconds_per_timestep: int) -> List[int]:
    """Tranform LPG timesteps (list of integers) from local time to UTC."""
    timeshifts = pytz.timezone("Europe/Berlin")._utc_transition_times  # type: ignore # pylint: disable=W0212
    timeshifts = [elem for elem in timeshifts if elem.year == year]
    steps_per_hour = int(3600 / seconds_per_timestep)
    timeshift1_as_step = (
        int((timeshifts[0] - dt.datetime(year=year, month=1, day=1)).seconds / seconds_per_timestep) - 1
    )
    timeshift2_as_step = (
        int((timeshifts[1] - dt.datetime(year=year, month=1, day=1)).seconds / seconds_per_timestep) - 1
    )

    data_utc = []
    for elem in data:
        if elem < timeshift1_as_step or elem > timeshift2_as_step:
            data_utc.append(elem - steps_per_hour)
        else:
            data_utc.append(elem - 2 * steps_per_hour)
    return data


def convert_lpg_data_to_utc(data: pd.DataFrame, year: int) -> pd.DataFrame:
    """Transform LPG data from local time (not having explicit time shifts) to UTC."""
    # convert Time information to pandas datetime and make it to index
    data.index = pd.DatetimeIndex(pd.to_datetime(data["Time"]))
    lastdate = data.index[-1]

    # find out time shifts of selected year
    timeshifts = pytz.timezone("Europe/Berlin")._utc_transition_times  # type: ignore # pylint: disable=W0212
    timeshifts = [elem for elem in timeshifts if elem.year == year]

    # delete hour in spring if neceary:
    if lastdate > timeshifts[0]:
        indices_of_additional_hour_in_spring = data.loc[
            timeshifts[0]
            + dt.timedelta(seconds=3600) : timeshifts[0]  # noqa: E203
            + dt.timedelta(seconds=60 * (60 + 59))
        ].index

        data.drop(index=indices_of_additional_hour_in_spring, inplace=True)

    # add hour in autumn if necesary
    if lastdate > timeshifts[1]:
        additional_hours_in_autumn = data.loc[
            timeshifts[1]
            + dt.timedelta(seconds=3600) : timeshifts[1]  # noqa: E203
            + dt.timedelta(seconds=60 * (60 + 59))
        ]
        data = pd.concat([data, additional_hours_in_autumn])
        data.sort_index(inplace=True)

    # delete hour at beginning
    data = data[data.index >= dt.datetime(year=year, month=1, day=1, hour=1)]

    # add hour at end
    last_hour = data[
        data.index >= dt.datetime(year=year, month=lastdate.month, day=lastdate.day, hour=23)  # type: ignore
    ]
    data = pd.concat([data, last_hour])  # type: ignore

    # make integer index again, paste new timestamp (UTC) and format
    data.index = pd.Index(list(range(len(data))))
    data["Time"] = pd.date_range(
        start=dt.datetime(year=year, month=1, day=1, hour=0),
        end=dt.datetime(year=year, month=lastdate.month, day=lastdate.day, hour=23, minute=59),  # type: ignore
        freq="min",
        tz="UTC",
    )
    data["Time"] = data["Time"].dt.strftime("%m/%d/%Y %H:%M")
    return data


def get_cache_file(
    component_key: str,
    parameter_class: Any,
    my_simulation_parameters: SimulationParameters,
    cache_dir_path: Optional[str] = None
) -> Tuple[bool, str]:  # noqa
    """Gets a cache path for a given parameter set.

    This will generate a file path based on any dataclass_json.
    It works by turning the class into a json string, hashing the string and then using that as filename.
    The idea is to have a unique file path for every possible configuration.
    """
    json_str = parameter_class.to_json()
    if cache_dir_path is None:
        cache_dir_path = my_simulation_parameters.cache_dir_path
    if my_simulation_parameters is None:
        raise ValueError("Simulation parameters was none.")
    simulation_parameter_str = my_simulation_parameters.get_unique_key()
    json_str = json_str + simulation_parameter_str
    if len(json_str) < 5:
        raise ValueError("Empty json detected for caching. This is a bug.")
    json_str_encoded = json_str.encode("utf-8")
    # Johanna Ganglbauer: python told me "TypeError: openssl_sha256() takes at most 1 argument (2 given)",
    # I removed the second input argument "usedforsecurity=False" and it works - maybe I need to update the hashlib package?
    sha_key = hashlib.sha256(json_str_encoded).hexdigest()
    filename = component_key + "_" + sha_key + ".cache"

    cache_absolute_filepath = os.path.join(cache_dir_path, filename)
    if not os.path.isdir(cache_dir_path):
        os.mkdir(cache_dir_path)
    if os.path.isfile(cache_absolute_filepath):
        return True, cache_absolute_filepath
    return False, cache_absolute_filepath


def load_export_load_profile_generator(target):  # noqa
    """Returns the paths for the SQL exported files from the Load Profile Generator."""
    targetpath = os.path.join(HISIMPATH["LoadProfileGenerator_export_directory"], target)
    if os.path.exists(targetpath):
        lpg_export_path = {
            "electric_vehicle": [
                os.path.join(targetpath, "Results.HH1.sqlite"),
                os.path.join(targetpath, "Results.General.sqlite"),
            ]
        }
        return lpg_export_path
    raise ValueError("Target export from Load Profile Generator does not exist")


def measure_execution_time(my_function):  # noqa
    """Utility function that works as decorator for measuring execution time."""

    @wraps(my_function)
    def function_wrapper_for_measuring_execution_time(*args, **kwargs):
        """Inner function for the time measuring utility decorator."""
        start = timer()
        result = my_function(*args, **kwargs)
        end = timer()
        diff = end - start
        log.profile(
            "Executing " + my_function.__module__ + "." + my_function.__name__ + " took " + f"{diff:1.2f}" + " seconds"
        )
        return result

    return function_wrapper_for_measuring_execution_time


def measure_memory_leak(my_function):  # noqa
    """Utility function that works as decorator for measuring execution time."""

    @wraps(my_function)
    def function_wrapper_for_measuring_memory_leak(*args, **kwargs):
        """Inner function for the time measuring utility decorator."""
        process = psutil.Process(os.getpid())
        rss_by_psutil_start = process.memory_info().rss / (1024 * 1024)
        result = my_function(*args, **kwargs)
        rss_by_psutil_end = process.memory_info().rss / (1024 * 1024)
        gc.collect()
        diff = rss_by_psutil_end - rss_by_psutil_start
        log.trace(
            "Executing " + my_function.__module__ + "." + my_function.__name__ + " leaked " + f"{diff:1.2f}" + " MB"
        )
        return result

    return function_wrapper_for_measuring_memory_leak


def measure_memory_leak_with_error(my_function):  # noqa
    """Utility function that works as decorator for measuring execution time."""

    @wraps(my_function)
    def function_wrapper_for_measuring_memory_leak(*args, **kwargs):
        """Inner function for the time measuring utility decorator."""
        process = psutil.Process(os.getpid())
        rss_by_psutil_start = process.memory_info().rss / (1024 * 1024)
        result = my_function(*args, **kwargs)
        rss_by_psutil_end = process.memory_info().rss / (1024 * 1024)
        gc.collect()
        diff = rss_by_psutil_end - rss_by_psutil_start
        log.information(
            "Executing " + my_function.__module__ + "." + my_function.__name__ + " leaked " + f"{diff:1.2f}" + " MB"
        )
        if diff > 100:
            raise ValueError("Lost over 100MB of memory during the function call")
        return result

    return function_wrapper_for_measuring_memory_leak


def deprecated(message):
    """Decorator for marking a function as deprecated."""

    def deprecated_decorator(func):
        """Decorator."""

        def deprecated_func(*args, **kwargs):
            """Core function."""
            log.warning(f"{func.__name__} is a deprecated function. {message}")
            return func(*args, **kwargs)

        return deprecated_func

    return deprecated_decorator


def rsetattr(obj, attr, val):
    """Recursive setattr for multi level attributes like `obj.attribute.subattribute`."""
    pre, _, post = attr.rpartition(".")
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


def rgetattr(obj, attr, *args):
    """Recursive getattr for multi level attributes like `obj.attribute.subattribute`."""

    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return freduce(_getattr, [obj] + attr.split("."))


def rhasattr(obj, attr):
    """Recursive hasattr for multi level attributes like `obj.attribute.subattribute`."""
    pre, _, post = attr.rpartition(".")
    return hasattr(rgetattr(obj, pre) if pre else obj, post)


def set_attributes_of_dataclass_from_dict(dataclass_, dict_, nested=None):
    """Set values in a Dataclass from a dictionary."""
    for key, value in dict_.items():
        if nested:
            path_list = nested + [key]
        else:
            path_list = [key]
        if isinstance(value, dict):
            set_attributes_of_dataclass_from_dict(dataclass_, value, path_list)
        else:
            attribute = ".".join(path_list)
            if rhasattr(dataclass_, attribute):
                rsetattr(dataclass_, attribute, value)
            else:
                raise AttributeError(
                    f"""Attribute `{attribute}` from JSON cannot be found
                    in `{dataclass_.__class__.__name__}`."""
                )


def get_environment_variable(key: str, default: Optional[str] = None) -> str:
    """Get environment variable. Raise error if variable not found."""
    value = os.getenv(key, default)
    if not value:
        raise ValueError(
            f"""Could not determine value of environment variable: {key}.
                         Make sure to set it in an `.env` file inside the HiSim root folder
                         or somewhere within your system environment."""
        )
    return value


class InstanceCounterMeta(type):
    """Metaclass to make instance counter not share count with descendants."""

    ids: itertools.count = itertools.count(1)


@dataclass
class InstanceCounter(metaclass=InstanceCounterMeta):
    """Mixin to add automatic ID generation."""

    def __post_init__(self, reset=False):
        """Runs after initialization of the dataclass."""
        if reset:
            self.instance_id = 0
        else:
            self.instance_id = next(self.__class__.ids)
        if self.instance_id > 1e3:
            raise RuntimeError(
                f"""Too many instances of {self.__class__}.
                Consider using a more performant or simpler type (e.g. int, float)."""
            )



"""Example Transformer."""

# clean

# Import packages from standard library or the environment e.g. pandas, numpy etc.
import os
from typing import List, Any
import json
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from scipy.interpolate import interp1d
import numpy as np

# Import modules from HiSim
from hisim.component import SingleTimeStepValues, ComponentInput, ComponentOutput, DisplayConfig
from hisim import loadtypes as lt
from hisim import utils
from hisim.simulationparameters import SimulationParameters

# from hisim.sim_repository_singleton import SingletonSimRepository, SingletonDictKeyEnum

from hisim import (
    component as cp,
)

__authors__ = "Franz Oldopp"
__copyright__ = "Copyright 2023, FZJ-IEK-3"
__credits__ = ["Franz Oldopp"]
__license__ = "-"
__version__ = "1.0"
__maintainer__ = "Franz Oldopp"
__status__ = "development"


@dataclass_json
@dataclass
class FuelCellConfig(cp.ConfigBase):
    """Configuration of the Example Transformer."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return FuelCell.get_full_classname()

    building_name: str
    # parameter_string: str
    # my_simulation_parameters: SimulationParameters
    name: str
    type: str
    nom_output: float  # [kW]
    max_output: float  # [kW]
    min_output: float  # [kW]
    nom_h2_flow_rate: float  # [m^3/h]
    faraday_eff: float
    i_cell_nom: float
    ramp_up_rate: float  # [%/s]
    ramp_down_rate: float  # [%/s]
    # H_s_h2 = 33.33 #kWh/kg

    @classmethod
    def get_default_pem_fuel_cell_config(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Gets a default PEM Eletrolyzer."""
        return FuelCellConfig(
            building_name=building_name,
            name="PEM_Fuel_Cell",
            type="PEM",
            nom_output=100.0,  # [kW]
            max_output=110.0,  # [kW]
            min_output=10.0,  # [kW]
            nom_h2_flow_rate=65.64,  # [m^3/h]
            faraday_eff=0.90,
            i_cell_nom=0.52,
            ramp_up_rate=0.1,  # [%/s]
            ramp_down_rate=0.2,  # [%/s]
            # H_s_h2 = 33.33,
        )

    @staticmethod
    def read_config(fuel_cell_name):
        """Read config."""
        config_file = os.path.join(utils.HISIMPATH["inputs"], "fuel_cell_manufacturer_config.json")
        with open(config_file, "r", encoding="utf-8") as json_file:
            data = json.load(json_file)
            return data.get("Fuel Cell variants", {}).get(fuel_cell_name, {})

    @classmethod
    def config_fuel_cell(
        cls,
        fuel_cell_name: str,
        building_name: str = "BUI1",
    ) -> Any:
        """Get config of fuel cell."""
        config_json = cls.read_config(fuel_cell_name)
        config = FuelCellConfig(
            building_name=building_name,
            name="FuelCell",  # config_json.get("name", "")
            type=config_json.get("type", ""),
            nom_output=config_json.get("nom_output", 0.0),
            max_output=config_json.get("max_output", 0.0),
            min_output=config_json.get("min_output", 0.0),
            nom_h2_flow_rate=config_json.get("nom_h2_flow_rate", 0.0),
            faraday_eff=config_json.get("faraday_eff", 0.0),
            i_cell_nom=config_json.get("i_cell_nom", 0.0),
            ramp_up_rate=config_json.get("ramp_up_rate", 0.0),
            ramp_down_rate=config_json.get("ramp_down_rate", 0.0),
        )
        return config


class FuelCell(cp.Component):
    """The Example Transformer class.

    It is used to modify input values and return them as new output values.

    Parameters
    ----------
    name : str
        Name of the electrolyzer variant.

    loadtype : lt.LoadTypes
        A :py:class:`~hisim.loadtypes.LoadTypes` object that represents
        the type of the loaded data.

    unit : lt.Units
        A :py:class:`~hisim.loadtypes.Units` object that represents
        the unit of the loaded data.

    min_load_electrolyzer : float
        Minimum capable load of thr electrolyzer.

    max_load_electrolyzer : float
        Maximum capable load of thr electrolyzer.

    """

    # Inputs
    DemandProfile = "DemandProfile"
    ControlSignal = "ControlSignal"

    # Outputs
    FuelCellState = "FuelCellState"
    WaterflowOutput = "WaterflowOutput"
    PowerOutput = "PowerOutput"
    TotalEnergyProduced = "TotalEnergyProduced"

    HydrogenDemand = "HydrogenDemand"
    OxygenDemand = "OxygenDemand"
    CurrentHydrogenFlowRate = "CurrentHydrogenFlowRate"
    CurrentOxygenFlowRate = "CurrentOxygenFlowRate"
    CurrentWaterFlowRate = "CurrentWaterFlowRate"
    TotalHydrogenConsumed = "TotalHydrogenConsumed"
    TotalOxygenConsumed = "TotalOxygenConsumed"
    TotalWaterProduced = "TotalWaterProduced"
    NumberofCycles = "NumberofCycles"
    TotalRampUpTime = "TotalRampUpTime"
    TotalRampDownTime = "TotalRampDownTime"
    CurrentEfficiency = "CurrentEfficiency"
    OperatingTime = "OperatingTime"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: FuelCellConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ):
        """Constructs all the neccessary attributes."""
        self.fuelcellconfig = config

        self.technology_type = config.type
        self.nom_output = config.nom_output
        self.max_output = config.max_output
        self.min_output = config.min_output
        self.nom_h2_flow_rate = config.nom_h2_flow_rate
        self.faraday_eff = config.faraday_eff
        self.i_cell_nom = config.i_cell_nom
        self.ramp_up_rate = config.ramp_up_rate
        self.ramp_down_rate = config.ramp_down_rate

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        # =================================================================================================================================
        # Input channels
        self.demand_profile_target: ComponentInput = self.add_input(
            self.fuelcellconfig.name,
            FuelCell.DemandProfile,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KILOWATT,
            True,
        )

        self.control_signal: ComponentInput = self.add_input(
            self.fuelcellconfig.name,
            FuelCell.ControlSignal,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            True,
        )
        # =================================================================================================================================
        # Output channels
        # Set state output
        self.fuel_cell_state: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.FuelCellState,
            lt.LoadTypes.ACTIVATION,
            lt.Units.ANY,
            output_description="Current state of the fuel cell",
        )

        self.total_energy_produced: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.TotalEnergyProduced,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KWH,
            output_description="Total energy produced for demand",
        )

        self.produced_water: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.WaterflowOutput,
            lt.LoadTypes.WATER,
            lt.Units.ANY,
            output_description="Current water flow rate",
        )

        self.current_power_output: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.PowerOutput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,  # for hosuehold use case
            postprocessing_flag=[
                lt.InandOutputType.ELECTRICITY_PRODUCTION,
                lt.ComponentType.FUEL_CELL,
            ],
            output_description="Current power output",
        )

        self.current_hydrogen_demand: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.HydrogenDemand,
            lt.LoadTypes.HYDROGEN,
            lt.Units.ANY,
            output_description="Current hydrogen demand",
        )

        self.number_cycles: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.NumberofCycles,
            lt.LoadTypes.ANY,
            lt.Units.ANY,
            output_description="Total number of activation cycles",
        )

        # Set total ramp-up time output
        self.total_ramp_up_time: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.TotalRampUpTime,
            lt.LoadTypes.TIME,
            lt.Units.SECONDS,
            output_description="Total ramp-up time",
        )

        # Set total ramp-down time output
        self.total_ramp_down_time: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.TotalRampDownTime,
            lt.LoadTypes.TIME,
            lt.Units.SECONDS,
            output_description="Total ramp-down time",
        )

        # current hydrogen output
        self.hydrogen_flow_rate: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.CurrentHydrogenFlowRate,
            lt.LoadTypes.HYDROGEN,
            lt.Units.KG_PER_SEC,
            output_description="Current hydrogen flow rate",
        )
        # Total hydrogen produced
        self.total_hydrogen: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.TotalHydrogenConsumed,
            lt.LoadTypes.HYDROGEN,
            lt.Units.KG,
            output_description="Total hydrogen produced during simulation time",
        )
        # current oxygen output
        self.oxygen_flow_rate: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.CurrentOxygenFlowRate,
            lt.LoadTypes.OXYGEN,
            lt.Units.KG_PER_SEC,
            output_description="Current oxygen flow rate",
        )
        # Total oxygen produced
        self.total_oxygen: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.TotalOxygenConsumed,
            lt.LoadTypes.OXYGEN,
            lt.Units.KG,
            output_description="Total oxygen produced during simulation time",
        )
        # current water demand
        self.water_flow_rate: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.CurrentWaterFlowRate,
            lt.LoadTypes.WATER,
            lt.Units.KG_PER_SEC,
            output_description="Current water flow rate",
        )
        # Total water demand
        self.total_water: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.TotalWaterProduced,
            lt.LoadTypes.WATER,
            lt.Units.KG,
            output_description="Total water demand during simulation time",
        )
        # Current efficiency
        self.current_efficiency_state: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.CurrentEfficiency,
            lt.LoadTypes.ANY,
            lt.Units.PERCENT,
            output_description="Current efficiency based on the efficiency curve",
        )

        # Total operating time
        self.operating_time: ComponentOutput = self.add_output(
            self.fuelcellconfig.name,
            FuelCell.OperatingTime,
            lt.LoadTypes.TIME,
            lt.Units.HOURS,
            output_description="Total time the electorlyzer is operating (on)",
        )

        # =================================================================================================================================
        # Transfer and storage of states
        self.current_power_state = 0.0
        self.total_ramp_up_count_state = 0.0
        self.total_ramp_down_count_state = 0.0
        self.total_warm_start_count = 0.0
        self.total_cold_start_count = 0.0
        self.total_warm_start_cycles = 0
        self.total_cold_start_cycles = 0
        self.current_warm_start_count = 0.0
        self.current_cold_start_count = 0.0
        self.total_hydrogen_consumed = 0.0
        self.total_oxygen_consumed = 0.0
        self.total_water_produced = 0.0
        self.total_operating_time = 0.0
        self.total_energy = 0.0

        self.current_power_state_previous = self.current_power_state
        self.total_ramp_up_count_state_previous = self.total_ramp_up_count_state
        self.total_ramp_down_count_state_previous = self.total_ramp_down_count_state
        self.total_warm_start_count_previous = self.total_warm_start_count
        self.total_cold_start_count_previous = self.total_cold_start_count
        self.total_warm_start_cycles_previous = self.total_warm_start_cycles
        self.total_cold_start_cycles_previous = self.total_cold_start_cycles
        self.current_warm_start_count_previous = self.current_warm_start_count
        self.current_cold_start_count_previous = self.current_cold_start_count
        self.total_hydrogen_consumed_previous = self.total_hydrogen_consumed
        self.total_oxygen_consumed_previous = self.total_oxygen_consumed
        self.total_water_produced_previous = self.total_water_produced
        self.total_operating_time_previous = self.total_operating_time
        self.total_energy_previous = self.total_energy

    @staticmethod
    def spec_el_stack_consumption_and_polarization_data_config(
        fuel_cell_type, nominal_power, h2_flow_rate, faraday_eff, i_cell_nom
    ):
        """Polarization curve data is provided corresponding to the used fuel cell technology.

        Following this, the auxiliary power of the system and the cell volatge is calculated,
        based on the nominal current density.
        """
        # Load data from the JSON file
        data_file = os.path.join(utils.HISIMPATH["inputs"], "polarization_curve_data_fc.json")
        with open(data_file, "r", encoding="utf-8") as file:
            data = json.load(file)

        # Check if the provided technology is valid
        if fuel_cell_type not in data:
            raise ValueError(
                f"{fuel_cell_type} is invalid technology. Supported technologies are: {', '.join(data.keys())}"
            )

        # Extract the x and y data points for the selected technology
        i_cell = data[fuel_cell_type]["i_cell"]
        u_cell = data[fuel_cell_type]["U_cell"]

        # constants
        f_constant = 96485  # C/mol
        m_h2 = 2.01588  # g/mol

        # from nom_current_density to aux_power
        spec_el_stack_consumption_nom = faraday_eff * (np.array(u_cell) * (2 * f_constant)) / (m_h2 * 3600)  # kWh/kg
        spec_el_consumption_stack = np.interp(i_cell_nom, i_cell, spec_el_stack_consumption_nom)

        # calculating aux_power
        aux_power = -nominal_power + (
            spec_el_consumption_stack * h2_flow_rate
        )  # might needs to be set to a constant value

        # interpolarization function
        u_cell_nom = np.interp(i_cell_nom, i_cell, u_cell)  # V

        return i_cell, u_cell, i_cell_nom, u_cell_nom, aux_power

    def h2_consumption_rate(
        self,
        i_cell_nom,
        u_cell_nom,
        nominal_power,
        min_power,
        i_cell,
        u_cell,
        h2_flow_rate,
        aux_power,
        current_power,
        state,
    ):
        """H2 consumption rate.

        Based on the polarisation curve, the spec. electricity demand and
        the current load, the H2 demand and the spec. H2 demand rate
        is calculated.
        """
        nominal_power_density = i_cell_nom * u_cell_nom  # W/cm²

        h2_consumption_rate = np.array(i_cell) / i_cell_nom * h2_flow_rate  # kg/h

        p_cell = np.array(i_cell) * np.array(u_cell)  # W/cm²

        stack_power = p_cell / nominal_power_density * nominal_power

        # Calculates system_power from stack power
        system_power = stack_power - aux_power

        interp_function_h2_consumption_rate = interp1d(system_power, h2_consumption_rate, kind="quadratic")

        negative_indices = np.where(system_power < 0.13081412540092785)

        # Entfernen Sie die negativen Werte aus beiden Listen
        filtered_system_power = np.delete(system_power, negative_indices)
        filtered_system_power[0] = 0.0
        filtered_h2_consumption_rate = np.delete(h2_consumption_rate, negative_indices)

        spec_h2_consumption_rate = (
            filtered_system_power / filtered_h2_consumption_rate
        )  # m³/kWh (proportional to the system efficiency)

        interp_function_spec_h2_demand_rate = interp1d(
            filtered_system_power, spec_h2_consumption_rate, kind="quadratic"
        )

        if state == 1 and current_power > min_power:
            # Only consume hydrogen if the system is "on"
            current_h2_demenad_rate = float(interp_function_h2_consumption_rate(current_power))

            current_spec_h2_demand_rate = float(interp_function_spec_h2_demand_rate(current_power))
            current_eff = current_spec_h2_demand_rate / 33.33  # LHV H2 33.33 kWh/kg
        elif state == 0 and current_power >= min_power:
            # Only consume hydrogen if the system is "on"
            current_h2_demenad_rate = float(interp_function_h2_consumption_rate(current_power))

            current_eff = 0.0

        else:
            # No hydrogen consumption if the system is in "standby" or "off"
            current_h2_demenad_rate = 0.0
            current_eff = 0.0

        return current_h2_demenad_rate, current_eff

    def oxygen_demand(self, current_h2_demenad_rate):
        """Oxygen demand.

        Returns the demand flow rate of oxygen,
        based on the current hydrogen flow rate.
        """
        m_o2 = 31.9988
        m_h2 = 2.01588
        m_dot_o2 = (m_o2 / m_h2) * 0.5 * current_h2_demenad_rate  # Kurzweil (2018) - Elektrolyse von Wasser
        return m_dot_o2

    def water_produced(self, current_h2_demenad_rate):
        """Water produced.

        Returns the produced water flow rate,
        based on the current hydrogen flow rate.
        """
        m_h2o = 18.01528
        m_h2 = 2.01588
        m_dot_h2o = (m_h2o / m_h2) * current_h2_demenad_rate  # Kurzweil (2018) - Elektrolyse von Wasser
        return m_dot_h2o

    def i_save_state(self) -> None:
        """Saves the current state."""
        self.current_power_state_previous = self.current_power_state
        self.total_ramp_up_count_state_previous = self.total_ramp_up_count_state
        self.total_ramp_down_count_state_previous = self.total_ramp_down_count_state
        self.total_warm_start_count_previous = self.total_warm_start_count
        self.total_cold_start_count_previous = self.total_cold_start_count
        self.total_warm_start_cycles_previous = self.total_warm_start_cycles
        self.total_cold_start_cycles_previous = self.total_cold_start_cycles
        self.current_warm_start_count_previous = self.current_warm_start_count
        self.current_cold_start_count_previous = self.current_cold_start_count
        self.total_hydrogen_consumed_previous = self.total_hydrogen_consumed
        self.total_oxygen_consumed_previous = self.total_oxygen_consumed
        self.total_water_produced_previous = self.total_water_produced
        self.total_operating_time_previous = self.total_operating_time
        self.total_energy_previous = self.total_energy

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_restore_state(self) -> None:
        """Restores previous state."""
        self.current_power_state = self.current_power_state_previous
        self.total_ramp_up_count_state = self.total_ramp_up_count_state_previous
        self.total_ramp_down_count_state = self.total_ramp_down_count_state_previous
        self.total_warm_start_count = self.total_warm_start_count_previous
        self.total_cold_start_count = self.total_cold_start_count_previous
        self.total_warm_start_cycles = self.total_warm_start_cycles_previous
        self.total_cold_start_cycles = self.total_cold_start_cycles_previous
        self.current_warm_start_count = self.current_warm_start_count_previous
        self.current_cold_start_count = self.current_cold_start_count_previous
        self.total_hydrogen_consumed = self.total_hydrogen_consumed_previous
        self.total_oxygen_consumed = self.total_oxygen_consumed_previous
        self.total_water_produced = self.total_water_produced_previous
        self.total_operating_time = self.total_operating_time_previous
        self.total_energy = self.total_energy_previous

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the component."""
        if force_convergence:
            return

        # Variables
        seconds_per_timestep = self.my_simulation_parameters.seconds_per_timestep  # [s/timestep]

        power_demand = stsv.get_input_value(self.demand_profile_target)
        # print("power_demand: ", power_demand)
        state = stsv.get_input_value(self.control_signal)

        # ramp up per timestep calculation
        ramp_up_per_timestep = self.nom_output * self.ramp_up_rate * seconds_per_timestep
        total_ramp_up_per_timestep = ramp_up_per_timestep

        # ramp down per timestep calculation
        ramp_down_per_timestep = self.nom_output * self.ramp_down_rate * seconds_per_timestep
        total_ramp_down_per_timestep = ramp_down_per_timestep

        # calculating the current power demand based on the previous state
        new_target = abs(power_demand - self.current_power_state)
        if state == 1:
            self.total_operating_time += seconds_per_timestep / 3600
            """the ramping process"""
            if new_target == self.nom_output:
                # print("punkt 2")
                self.current_power_state = self.nom_output
                self.total_ramp_up_count_state += 0
                self.total_ramp_down_count_state += 0
            if new_target == 0:
                # print("punkt 3")
                self.current_power_state = self.current_power_state
                self.total_ramp_up_count_state += 0
                self.total_ramp_down_count_state += 0

            # Ramping up
            if new_target >= total_ramp_up_per_timestep and self.current_power_state < power_demand:
                # print("punkt 4")
                self.total_ramp_up_count_state += seconds_per_timestep
                self.current_power_state += total_ramp_up_per_timestep
                # print("self.current_power_state nach punkt 4: ", self.current_power_state)
            elif self.current_power_state < power_demand and new_target < total_ramp_up_per_timestep:
                # print("punkt 5")
                percentage_ramp_up_per_timestep = new_target / total_ramp_up_per_timestep
                self.total_ramp_up_count_state += percentage_ramp_up_per_timestep * seconds_per_timestep
                if self.current_power_state == 0:
                    # print("punkt 6")
                    self.current_power_state += power_demand
                else:
                    # print("punkt 7")
                    self.current_power_state += new_target

            # Ramping down
            elif total_ramp_down_per_timestep <= new_target and power_demand < self.current_power_state:
                # print("punkt 8")
                self.total_ramp_down_count_state += seconds_per_timestep
                self.current_power_state -= new_target

            elif power_demand < self.current_power_state and new_target < total_ramp_down_per_timestep:
                # print("punkt 9")
                percentage_ramp_down_per_timestep = new_target / total_ramp_down_per_timestep
                self.total_ramp_down_count_state += percentage_ramp_down_per_timestep * seconds_per_timestep
                self.current_power_state -= new_target

        elif state == 0:
            # print("punkt 10")
            self.total_ramp_up_count_state += 0
            self.total_ramp_down_count_state += 0
            self.current_power_state = 0.0

        elif state == -1:
            # print("punkt 11")
            self.total_ramp_up_count_state += 0
            self.total_ramp_down_count_state += 0
            self.current_power_state = 0.0

        # Applying polarization curve data
        (
            i_cell,
            u_cell,
            i_cell_nom,
            u_cell_nom,
            aux_power,
        ) = self.spec_el_stack_consumption_and_polarization_data_config(
            self.technology_type,
            self.nom_output,
            self.nom_h2_flow_rate,
            self.faraday_eff,
            self.i_cell_nom,
        )
        # Current hydrogen prduction and specific hydrogen production rate

        # self.current_spec_h2_demand_rate,
        current_h2_demand_rate, current_eff = self.h2_consumption_rate(
            i_cell_nom,
            u_cell_nom,
            self.nom_output,
            self.min_output,
            i_cell,
            u_cell,
            self.nom_h2_flow_rate,
            aux_power,
            self.current_power_state,
            state,
        )
        # Current oxygen and water flow rate
        current_flow_rate_oxygen = self.oxygen_demand(current_h2_demand_rate)
        current_flow_rate_water = self.water_produced(current_h2_demand_rate)
        # Calculating total amount of hydrogen, oxygen and water
        total_hydrogen_consumed_in_timestep = current_h2_demand_rate * (seconds_per_timestep / 3600)
        self.total_hydrogen_consumed += total_hydrogen_consumed_in_timestep
        total_oxygen_consumed_in_timestep = current_flow_rate_oxygen * (seconds_per_timestep / 3600)
        self.total_oxygen_consumed += total_oxygen_consumed_in_timestep
        total_water_produced_in_timestep = current_flow_rate_water * (seconds_per_timestep / 3600)
        self.total_water_produced += total_water_produced_in_timestep

        self.total_energy += self.current_power_state * (seconds_per_timestep / 3600)
        # Initializing outputs
        stsv.set_output_value(self.hydrogen_flow_rate, current_h2_demand_rate)
        stsv.set_output_value(self.oxygen_flow_rate, current_flow_rate_oxygen)
        stsv.set_output_value(self.water_flow_rate, current_flow_rate_water)
        stsv.set_output_value(self.fuel_cell_state, state)
        stsv.set_output_value(
            self.current_power_output, (self.current_power_state * 1000)
        )  # transform kW to WATT for EMS
        stsv.set_output_value(self.total_energy_produced, self.total_energy)
        stsv.set_output_value(self.total_ramp_up_time, self.total_ramp_up_count_state)
        stsv.set_output_value(self.total_ramp_down_time, self.total_ramp_down_count_state)
        stsv.set_output_value(self.total_hydrogen, self.total_hydrogen_consumed)
        stsv.set_output_value(self.total_oxygen, self.total_oxygen_consumed)
        stsv.set_output_value(self.total_water, self.total_water_produced)
        stsv.set_output_value(self.operating_time, self.total_operating_time)
        stsv.set_output_value(self.current_efficiency_state, current_eff)

    def write_to_report(self) -> List[str]:
        """Writes a report."""
        lines = []
        for config_string in self.fuelcellconfig.get_string_dict():
            lines.append(config_string)
        lines.append("Component Name" + str(self.component_name))
        lines.append("Total operating time during simulation: " + str(self.total_operating_time) + " [h]")
        lines.append("Total hydrogen consumed during simulation: " + str(self.total_hydrogen_consumed) + " [kg]")
        lines.append("Total oxygen consumed during simulation: " + str(self.total_oxygen_consumed) + " [kg]")
        lines.append("Total water demand during simulation: " + str(self.total_water_produced) + " [kg]")
        lines.append("Total energy produced during simulation: " + str(self.total_energy) + " [kWh]")
        lines.append("Total ramp-up time during simulation: " + str(self.total_ramp_up_count_state) + " [kg]")
        lines.append("Total ramp-down time during simulation: " + str(self.total_ramp_down_count_state) + " [kg]")
        return lines



"""Electricity meter module should replace the sumbuilder. """

# clean
from dataclasses import dataclass
from typing import List

import pandas as pd
from dataclasses_json import dataclass_json

from hisim import component as cp
from hisim import dynamic_component
from hisim import loadtypes as lt
from hisim.component import ComponentInput, OpexCostDataClass, CapexCostDataClass
from hisim.components.configuration import EmissionFactorsAndCostsForFuelsConfig
from hisim.dynamic_component import (
    DynamicComponent,
    DynamicConnectionInput,
    DynamicConnectionOutput,
    DynamicComponentConnection,
)
from hisim.simulationparameters import SimulationParameters
from hisim.postprocessing.kpi_computation.kpi_structure import KpiEntry, KpiTagEnumClass


@dataclass_json
@dataclass
class ElectricityMeterConfig(cp.ConfigBase):
    """Electricity Meter Config."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return ElectricityMeter.get_full_classname()

    building_name: str
    name: str

    @classmethod
    def get_electricity_meter_default_config(
        cls,
        name: str = "ElectricityMeter",
        building_name: str = "BUI1",
    ) -> "ElectricityMeterConfig":
        """Gets a default ElectricityMeter."""
        return ElectricityMeterConfig(
            building_name=building_name,
            name=name,
        )


class ElectricityMeter(DynamicComponent):
    """Electricity meter class.

    It calculates the electricity production and consumption dynamically for all components.
    """

    # Outputs
    ElectricityAvailable = "ElectricityAvailable"
    ElectricityToAndFromGrid = "ElectricityToAndFromGrid"
    ElectricityToGrid = "ElectricityToGrid"
    ElectricityFromGrid = "ElectricityFromGrid"
    ElectricityConsumption = "ElectricityConsumption"
    ElectricityProduction = "ElectricityProduction"
    CumulativeConsumption = "CumulativeConsumption"
    CumulativeProduction = "CumulativeProduction"
    ElectricityToGridInWatt = "ElectricityToGridInWatt"
    ElectricityFromGridInWatt = "ElectricityFromGridInWatt"
    ElectricityProductionInWatt = "ElectricityProductionInWatt"
    ElectricityConsumptionInWatt = "ElectricityConsumptionInWatt"
    SurplusUnusedFromBuildingEMSOutput = "SurplusUnusedFromBuildingEMSOutput"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: ElectricityMeterConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(display_in_webtool=True),
    ):
        """Initialize the component."""
        self.grid_energy_balancer_config = config
        self.name = self.grid_energy_balancer_config.name
        self.my_component_inputs: List[DynamicConnectionInput] = []
        self.my_component_outputs: List[DynamicConnectionOutput] = []
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            self.my_component_inputs,
            self.my_component_outputs,
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        self.production_inputs: List[ComponentInput] = []
        self.consumption_uncontrolled_inputs: List[ComponentInput] = []

        self.seconds_per_timestep = self.my_simulation_parameters.seconds_per_timestep
        # Component has states
        self.state = ElectricityMeterState(cumulative_production_in_watt_hour=0, cumulative_consumption_in_watt_hour=0)
        self.previous_state = self.state.self_copy()

        # Outputs
        self.electricity_to_grid_in_watt_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityToGridInWatt,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ElectricityToGridInWatt} will follow.",
        )
        self.electricity_from_grid_in_watt_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityFromGridInWatt,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ElectricityFromGridInWatt} will follow.",
        )
        self.electricity_production_in_watt_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityProductionInWatt,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ElectricityProductionInWatt} will follow.",
        )
        if any(word in config.building_name for word in lt.DistrictNames):
            self.surplus_electricity_unused_to_district_ems_from_building_ems_output: cp.ComponentOutput = self.add_output(
                object_name=self.component_name,
                field_name=self.SurplusUnusedFromBuildingEMSOutput,
                load_type=lt.LoadTypes.ELECTRICITY,
                unit=lt.Units.WATT,
                sankey_flow_direction=False,
                output_description=f"here a description for {self.SurplusUnusedFromBuildingEMSOutput} will follow.",
                postprocessing_flag=(
                    [
                        lt.InandOutputType.ELECTRICITY_PRODUCTION,
                        lt.ComponentType.BUILDINGS,
                        lt.OutputPostprocessingRules.DISPLAY_IN_WEBTOOL,
                    ]
                    if any(word in config.building_name for word in lt.DistrictNames)
                    else []
                ),
            )
        self.electricity_consumption_uncontrolled_in_watt_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityConsumptionInWatt,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ElectricityConsumptionInWatt} will follow.",
            postprocessing_flag=(
                [
                    lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED,
                    lt.ComponentType.BUILDINGS,
                    lt.OutputPostprocessingRules.DISPLAY_IN_WEBTOOL,
                ]
                if any(word in config.building_name for word in lt.DistrictNames)
                else []
            ),
        )
        self.electricity_available_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityAvailable,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ElectricityAvailable} will follow.",
        )
        self.electricity_to_and_from_grid_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityToAndFromGrid,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT_HOUR,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ElectricityToAndFromGrid} will follow.",
        )
        self.electricity_to_grid_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityToGrid,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT_HOUR,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ElectricityToGrid} will follow.",
            postprocessing_flag=[lt.OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
        )
        self.electricity_from_grid_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityFromGrid,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT_HOUR,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ElectricityFromGrid} will follow.",
            postprocessing_flag=[lt.OutputPostprocessingRules.DISPLAY_IN_WEBTOOL],
        )

        self.electricity_consumption_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityConsumption,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT_HOUR,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ElectricityConsumption} will follow.",
        )

        self.electricity_production_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityProduction,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT_HOUR,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ElectricityProduction} will follow.",
        )

        self.cumulative_electricity_consumption_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.CumulativeConsumption,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT_HOUR,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.CumulativeConsumption} will follow.",
        )

        self.cumulative_electricity_production_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.CumulativeProduction,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT_HOUR,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.CumulativeProduction} will follow.",
        )
        self.add_dynamic_default_connections(self.get_default_connections_from_utsp_occupancy())
        self.add_dynamic_default_connections(self.get_default_connections_from_pv_system())
        self.add_dynamic_default_connections(self.get_default_connections_from_dhw_heat_pump())
        self.add_dynamic_default_connections(self.get_default_connections_from_advanced_heat_pump())

    def get_default_connections_from_utsp_occupancy(
        self,
    ):
        """Get utsp occupancy default connections."""

        from hisim.components.loadprofilegenerator_utsp_connector import (  # pylint: disable=import-outside-toplevel
            UtspLpgConnector,
        )

        dynamic_connections = []
        occupancy_class_name = UtspLpgConnector.get_classname()
        dynamic_connections.append(
            dynamic_component.DynamicComponentConnection(
                source_component_class=UtspLpgConnector,
                source_class_name=occupancy_class_name,
                source_component_field_name=UtspLpgConnector.ElectricityOutput,
                source_load_type=lt.LoadTypes.ELECTRICITY,
                source_unit=lt.Units.WATT,
                source_tags=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED],
                source_weight=999,
            )
        )
        return dynamic_connections

    def get_default_connections_from_pv_system(
        self,
    ):
        """Get pv system default connections."""

        from hisim.components.generic_pv_system import PVSystem  # pylint: disable=import-outside-toplevel

        dynamic_connections = []
        pv_class_name = PVSystem.get_classname()
        dynamic_connections.append(
            DynamicComponentConnection(
                source_component_class=PVSystem,
                source_class_name=pv_class_name,
                source_component_field_name=PVSystem.ElectricityOutput,
                source_load_type=lt.LoadTypes.ELECTRICITY,
                source_unit=lt.Units.WATT,
                source_tags=[
                    lt.ComponentType.PV,
                    lt.InandOutputType.ELECTRICITY_PRODUCTION,
                ],
                source_weight=999,
            )
        )
        return dynamic_connections

    def get_default_connections_from_dhw_heat_pump(
        self,
    ):
        """Get dhw heat pump default connections."""

        from hisim.components.generic_heat_pump_modular import (  # pylint: disable=import-outside-toplevel
            ModularHeatPump,
        )

        dynamic_connections = []
        dhw_heat_pump_class_name = ModularHeatPump.get_classname()
        dynamic_connections.append(
            DynamicComponentConnection(
                source_component_class=ModularHeatPump,
                source_class_name=dhw_heat_pump_class_name,
                source_component_field_name=ModularHeatPump.ElectricityOutput,
                source_load_type=lt.LoadTypes.ELECTRICITY,
                source_unit=lt.Units.WATT,
                source_tags=[lt.ComponentType.HEAT_PUMP_DHW, lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED],
                source_weight=999,
            )
        )
        return dynamic_connections

    def get_default_connections_from_advanced_heat_pump(
        self,
    ):
        """Get advanced heat pump default connections."""

        from hisim.components.advanced_heat_pump_hplib import HeatPumpHplib  # pylint: disable=import-outside-toplevel

        dynamic_connections = []
        advanced_heat_pump_class_name = HeatPumpHplib.get_classname()
        dynamic_connections.append(
            DynamicComponentConnection(
                source_component_class=HeatPumpHplib,
                source_class_name=advanced_heat_pump_class_name,
                source_component_field_name=HeatPumpHplib.ElectricalInputPower,
                source_load_type=lt.LoadTypes.ELECTRICITY,
                source_unit=lt.Units.WATT,
                source_tags=[
                    lt.ComponentType.HEAT_PUMP_BUILDING,
                    lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED,
                ],
                source_weight=999,
            )
        )
        return dynamic_connections

    def write_to_report(self):
        """Writes relevant information to report."""
        return self.grid_energy_balancer_config.get_string_dict()

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.self_copy()

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = self.previous_state.self_copy()

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks values."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the grid energy balancer."""

        if timestep == 0:
            self.production_inputs = self.get_dynamic_inputs(tags=[lt.InandOutputType.ELECTRICITY_PRODUCTION])
            self.consumption_uncontrolled_inputs = self.get_dynamic_inputs(
                tags=[lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED]
            )

        # ELECTRICITY #

        # get sum of production and consumption for all inputs for each iteration
        production_in_watt = sum([stsv.get_input_value(component_input=elem) for elem in self.production_inputs])
        consumption_uncontrolled_in_watt = sum(
            [stsv.get_input_value(component_input=elem) for elem in self.consumption_uncontrolled_inputs]
        )

        if any(word in self.config.building_name for word in lt.DistrictNames):
            production_inputs_building = self.get_dynamic_inputs(tags=[lt.InandOutputType.ELECTRICITY_PRODUCTION, lt.ComponentType.BUILDINGS])

            building_electricity_surplus_unused = (
                sum([stsv.get_input_value(component_input=elem) for elem in production_inputs_building]))

            stsv.set_output_value(
                self.surplus_electricity_unused_to_district_ems_from_building_ems_output,
                building_electricity_surplus_unused,
            )
        # Production of Electricity positve sign
        # Consumption of Electricity negative sign
        difference_between_production_and_consumption_in_watt = production_in_watt - consumption_uncontrolled_in_watt

        # transform watt to watthour
        production_in_watt_hour = production_in_watt * self.seconds_per_timestep / 3600
        consumption_uncontrolled_in_watt_hour = consumption_uncontrolled_in_watt * self.seconds_per_timestep / 3600
        difference_between_production_and_consumption_in_watt_hour = (
            production_in_watt_hour - consumption_uncontrolled_in_watt_hour
        )

        # calculate cumulative production and consumption
        cumulative_production_in_watt_hour = self.state.cumulative_production_in_watt_hour + production_in_watt_hour
        cumulative_consumption_in_watt_hour = (
            self.state.cumulative_consumption_in_watt_hour + consumption_uncontrolled_in_watt_hour
        )

        # consumption is bigger than production -> electricity from grid is needed
        # change sign so that value becomes positive
        if difference_between_production_and_consumption_in_watt_hour < 0:
            electricity_from_grid_in_watt_hour = -difference_between_production_and_consumption_in_watt_hour
            electricity_to_grid_in_watt_hour = 0.0
        # production is bigger -> electricity can be fed into grid
        elif difference_between_production_and_consumption_in_watt_hour > 0:
            electricity_to_grid_in_watt_hour = difference_between_production_and_consumption_in_watt_hour
            electricity_from_grid_in_watt_hour = 0.0

        # difference between production and consumption is zero
        else:
            electricity_to_grid_in_watt_hour = 0.0
            electricity_from_grid_in_watt_hour = 0.0

        # set outputs
        stsv.set_output_value(
            self.electricity_to_grid_in_watt_channel,
            (
                difference_between_production_and_consumption_in_watt
                if difference_between_production_and_consumption_in_watt > 0
                else 0
            ),
        )
        stsv.set_output_value(
            self.electricity_from_grid_in_watt_channel,
            (
                -difference_between_production_and_consumption_in_watt
                if difference_between_production_and_consumption_in_watt < 0
                else 0
            ),
        )

        stsv.set_output_value(
            self.electricity_production_in_watt_channel,
            production_in_watt,
        )

        stsv.set_output_value(
            self.electricity_consumption_uncontrolled_in_watt_channel,
            consumption_uncontrolled_in_watt,
        )

        stsv.set_output_value(
            self.electricity_available_channel,
            difference_between_production_and_consumption_in_watt,
        )
        stsv.set_output_value(
            self.electricity_to_and_from_grid_channel,
            difference_between_production_and_consumption_in_watt_hour,
        )
        stsv.set_output_value(self.electricity_to_grid_channel, electricity_to_grid_in_watt_hour)
        stsv.set_output_value(self.electricity_from_grid_channel, electricity_from_grid_in_watt_hour)
        stsv.set_output_value(
            self.electricity_consumption_channel,
            consumption_uncontrolled_in_watt_hour,
        )

        stsv.set_output_value(
            self.electricity_production_channel,
            production_in_watt_hour,
        )

        stsv.set_output_value(
            self.cumulative_electricity_consumption_channel,
            cumulative_consumption_in_watt_hour,
        )

        stsv.set_output_value(
            self.cumulative_electricity_production_channel,
            cumulative_production_in_watt_hour,
        )

        self.state.cumulative_production_in_watt_hour = cumulative_production_in_watt_hour
        self.state.cumulative_consumption_in_watt_hour = cumulative_consumption_in_watt_hour

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        """Calculate OPEX costs, consisting of electricity costs and revenues."""
        total_energy_to_grid_in_kwh: float
        total_energy_from_grid_in_kwh: float

        for index, output in enumerate(all_outputs):
            if output.component_name == self.component_name:
                if output.field_name == self.ElectricityToGrid:
                    # Todo: check component name from system_setups: find another way of using the correct outputs
                    total_energy_to_grid_in_kwh = round(postprocessing_results.iloc[:, index].sum() * 1e-3, 2)

                elif output.field_name == self.ElectricityFromGrid:
                    total_energy_from_grid_in_kwh = round(postprocessing_results.iloc[:, index].sum() * 1e-3, 2)

        emissions_and_cost_factors = EmissionFactorsAndCostsForFuelsConfig.get_values_for_year(
            self.my_simulation_parameters.year
        )
        co2_per_unit = emissions_and_cost_factors.electricity_footprint_in_kg_per_kwh
        euro_per_unit = emissions_and_cost_factors.electricity_costs_in_euro_per_kwh
        revenue_euro_per_unit = emissions_and_cost_factors.electricity_to_grid_revenue_in_euro_per_kwh

        opex_cost_per_simulated_period_in_euro = (
            total_energy_from_grid_in_kwh * euro_per_unit - total_energy_to_grid_in_kwh * revenue_euro_per_unit
        )
        co2_per_simulated_period_in_kg = total_energy_from_grid_in_kwh * co2_per_unit
        opex_cost_data_class = OpexCostDataClass(
            opex_energy_cost_in_euro=opex_cost_per_simulated_period_in_euro,
            opex_maintenance_cost_in_euro=0,
            co2_footprint_in_kg=co2_per_simulated_period_in_kg,
            consumption_in_kwh=total_energy_from_grid_in_kwh,
            loadtype=lt.LoadTypes.ELECTRICITY,
            kpi_tag=KpiTagEnumClass.ELECTRICITY_METER
        )

        return opex_cost_data_class

    @staticmethod
    def get_cost_capex(config: ElectricityMeterConfig, simulation_parameters: SimulationParameters) -> CapexCostDataClass:  # pylint: disable=unused-argument
        """Returns investment cost, CO2 emissions and lifetime."""
        capex_cost_data_class = CapexCostDataClass.get_default_capex_cost_data_class()
        return capex_cost_data_class

    def get_component_kpi_entries(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        total_energy_from_grid_in_kwh: float
        total_energy_to_grid_in_kwh: float
        list_of_kpi_entries: List[KpiEntry] = []
        for index, output in enumerate(all_outputs):
            if output.component_name == self.component_name and output.load_type == lt.LoadTypes.ELECTRICITY:
                if output.field_name == self.ElectricityFromGrid:
                    total_energy_from_grid_in_kwh = postprocessing_results.iloc[:, index].sum() * 1e-3
                elif output.field_name == self.ElectricityToGrid:
                    total_energy_to_grid_in_kwh = postprocessing_results.iloc[:, index].sum() * 1e-3

        total_energy_from_grid_in_kwh_entry = KpiEntry(
            name="Total energy from grid",
            unit="kWh",
            value=total_energy_from_grid_in_kwh,
            tag=KpiTagEnumClass.ELECTRICITY_METER,
            description=self.component_name,
        )
        list_of_kpi_entries.append(total_energy_from_grid_in_kwh_entry)

        total_energy_to_grid_in_kwh_entry = KpiEntry(
            name="Total energy to grid",
            unit="kWh",
            value=total_energy_to_grid_in_kwh,
            tag=KpiTagEnumClass.ELECTRICITY_METER,
            description=self.component_name,
        )
        list_of_kpi_entries.append(total_energy_to_grid_in_kwh_entry)

        # get opex costs
        opex_costs = self.get_cost_opex(all_outputs=all_outputs, postprocessing_results=postprocessing_results)
        opex_costs_in_euro_entry = KpiEntry(
            name="Opex costs of electricity consumption from grid",
            unit="Euro",
            value=opex_costs.opex_energy_cost_in_euro,
            tag=KpiTagEnumClass.ELECTRICITY_METER,
            description=self.component_name,
        )
        list_of_kpi_entries.append(opex_costs_in_euro_entry)
        co2_footprint_in_kg_entry = KpiEntry(
            name="CO2 footprint of electricity consumption from grid",
            unit="kg",
            value=opex_costs.co2_footprint_in_kg,
            tag=KpiTagEnumClass.ELECTRICITY_METER,
            description=self.component_name,
        )
        list_of_kpi_entries.append(co2_footprint_in_kg_entry)

        return list_of_kpi_entries


@dataclass
class ElectricityMeterState:
    """ElectricityMeterState class."""

    cumulative_production_in_watt_hour: float
    cumulative_consumption_in_watt_hour: float

    def self_copy(
        self,
    ):
        """Copy the ElectricityMeterState."""
        return ElectricityMeterState(
            self.cumulative_production_in_watt_hour,
            self.cumulative_consumption_in_watt_hour,
        )



"""rSOC."""

# clean
import os
import json
from typing import Any
from dataclasses import dataclass
from dataclasses_json import dataclass_json
import numpy as np

# Import modules from HiSim
from hisim.component import SingleTimeStepValues, ComponentInput, ComponentOutput, DisplayConfig
from hisim import loadtypes as lt
from hisim import utils
from hisim.simulationparameters import SimulationParameters

from hisim import component as cp

__authors__ = "Franz Oldopp"
__copyright__ = "Copyright 2023, FZJ-IEK-3"
__credits__ = ["Franz Oldopp"]
__license__ = "-"
__version__ = "1.0"
__maintainer__ = "Franz Oldopp"
__status__ = "development"


@dataclass_json
@dataclass
class RsocConfig(cp.ConfigBase):
    """Configuration of the rSOC."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return Rsoc.get_full_classname()

    building_name: str
    name: str
    # SOEC
    nom_load_soec: float  # [kW]
    min_load_soec: float  # [kW]
    max_load_soec: float  # [kW]
    faraday_eff_soec: float
    ramp_up_rate_soec: float  # [%/s]
    ramp_down_rate_soec: float  # [%/s]
    # SOFC
    nom_power_sofc: float  # [kW]
    min_power_sofc: float  # [kW]
    max_power_sofc: float  # [kW]
    faraday_eff_sofc: float
    ramp_up_rate_sofc: float  # [%/s]
    ramp_down_rate_sofc: float  # [%/s]

    @staticmethod
    def read_config(rsoc_name):
        """Opens the according JSON-file, based on the rSOC_name."""

        config_file = os.path.join(utils.HISIMPATH["inputs"], "rSOC_manufacturer_config.json")
        with open(config_file, "r", encoding="utf-8") as json_file:
            data = json.load(json_file)
            return data.get("rSOC variants", {}).get(rsoc_name, {})

    @classmethod
    def config_rsoc(
        cls,
        rsoc_name: str,
        building_name: str = "BUI1",
    ) -> Any:
        """Initializes the config variables based on the JSON-file."""

        config_json = cls.read_config(rsoc_name)
        config = RsocConfig(
            building_name=building_name,
            name=rsoc_name,  # config_json.get("name", "")
            nom_load_soec=config_json.get("nom_load_soec", 0.0),
            min_load_soec=config_json.get("min_load_soec", 0.0),
            max_load_soec=config_json.get("max_load_soec", 0.0),
            faraday_eff_soec=config_json.get("faraday_eff_soec", 0.0),
            ramp_up_rate_soec=config_json.get("ramp_up_rate_soec", 0.0),
            ramp_down_rate_soec=config_json.get("ramp_down_rate_soec", 0.0),
            nom_power_sofc=config_json.get("nom_power_sofc", 0.0),
            min_power_sofc=config_json.get("min_power_sofc", 0.0),
            max_power_sofc=config_json.get("max_power_sofc", 0.0),
            faraday_eff_sofc=config_json.get("faraday_eff_sofc", 0.0),
            ramp_up_rate_sofc=config_json.get("ramp_up_rate_sofc", 0.0),
            ramp_down_rate_sofc=config_json.get("ramp_down_rate_sofc", 0.0),
        )
        return config


class Rsoc(cp.Component):
    """Rsoc component class."""

    # Inputs
    PowerInput = "PowerInput"
    RSOCInputState = "RSOCInputState"

    # SOEC Outputs
    SOECState = "SOECState"
    SOECCurrentLoad = "SOECCurrentLoadConsumed"  # current load regarding the ramp-up and -down process
    TotalEnergyConsumed = "TotalEnergyConsumed"

    SOECCurrentHydrogenFlowRate = "SOECCurrentHydrogenFlowRate"
    SOECCurrentOxygenFlowRate = "SOECCurrentOxygenFlowRate"
    SOECCurrentWaterFlowRate = "SOECCurrentWaterFlowRate"
    SOECTotalHydrogenProduced = "SOECTotalHydrogenProduced"
    SOECTotalOxygenProduced = "SOECTotalOxygenProduced"
    SOECTotalWaterDemand = "SOECTotalWaterDemand"

    SOECCurrentEfficiency = "SOECCurrentEfficiency"

    SOECTotalWarmStartTime = "SOECTotalWarmStartTime"
    SOECWarmStartCycles = "SOECWarmStartCycles"
    SOECTotalColdStartTime = "SOECTotalColdStartTime"
    SOECColdStartCycles = "SOECColdStartCycles"
    SOECTotalRampUpTime = "SOECTotalRampUpTime"
    SOECTotalRampDownTime = "SOECTotalRampDownTime"

    # SOFC Outputs
    SOFCState = "SOFCState"
    SOFCCurrentOutput = "SOFCCurrentPowerOutput"  # current load regarding the ramp-up and -down process
    TotalEnergyProduced = "TotalEnergyProduced"

    SOFCCurrentHydrogenFlowRate = "SOFCCurrentHydrogenFlowRate"
    SOFCCurrentOxygenFlowRate = "SOECCurrentOxygenFlowRate"
    SOFCCurrentWaterFlowRate = "SOECCurrentWaterFlowRate"
    SOFCTotalHydrogenConsumed = "SOFCTotalHydrogenConsumed"
    SOFCTotalOxygenConsumed = "SOFCTotalOxygenConsumed"
    SOFCTotalWaterProduced = "SOFCTotalWaterProduced"

    SOFCCurrentEfficiency = "SOFCCurrentEfficiency"

    SOFCTotalWarmStartTime = "SOECTotalWarmStartTime"
    SOFCWarmStartCycles = "SOECWarmStartCycles"
    SOFCTotalColdStartTime = "SOECTotalColdStartTime"
    SOFCColdStartCycles = "SOECColdStartCycles"
    SOFCTotalRampUpTime = "SOECTotalRampUpTime"
    SOFCTotalRampDownTime = "SOECTotalRampDownTime"

    RSOCOperatingTime = "rSOCOperatingTime"  # (withoutswitchingtimes)

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: RsocConfig,
        my_display_config: DisplayConfig = DisplayConfig(),
    ):
        """Constructs all the neccessary attributes."""
        self.rsocconfig = config

        self.name = config.name
        self.nom_load_soec = config.nom_load_soec
        self.min_load_soec = config.min_load_soec
        self.max_load_soec = config.max_load_soec
        self.faraday_eff_soec = config.faraday_eff_soec
        self.ramp_up_rate_soec = config.ramp_up_rate_soec
        self.ramp_down_rate_soec = config.ramp_down_rate_soec

        self.nom_power_sofc = config.nom_power_sofc
        self.min_power_sofc = config.min_power_sofc
        self.max_power_sofc = config.max_power_sofc
        self.faraday_eff_sofc = config.faraday_eff_sofc
        self.ramp_up_rate_sofc = config.ramp_up_rate_sofc
        self.ramp_down_rate_sofc = config.ramp_down_rate_sofc

        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )

        # =================================================================================================================================
        # Input channels
        self.power_input: ComponentInput = self.add_input(
            self.rsocconfig.name,
            Rsoc.PowerInput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KILOWATT,
            True,
        )

        # get the state from the controller
        self.input_state_rsoc: ComponentInput = self.add_input(
            self.rsocconfig.name,
            Rsoc.RSOCInputState,
            lt.LoadTypes.ACTIVATION,
            lt.Units.ANY,
            True,
        )

        # =================================================================================================================================
        # Output channels

        self.soec_current_efficiency_state: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.SOECCurrentEfficiency,
            lt.LoadTypes.ANY,
            lt.Units.PERCENT,
            output_description="Current efficiency based on the efficiency curve",
        )

        # current hydrogen output
        self.soec_hydrogen_flow_rate: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.SOECCurrentHydrogenFlowRate,
            lt.LoadTypes.HYDROGEN,
            lt.Units.KG_PER_SEC,
            output_description="Current hydrogen flow rate",
        )

        # current hydrogen consumption
        self.sofc_hydrogen_flow_rate: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.SOFCCurrentHydrogenFlowRate,
            lt.LoadTypes.HYDROGEN,
            lt.Units.KG_PER_SEC,
            output_description="Current hydrogen flow rate",
        )

        self.sofc_current_efficiency_state: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.SOFCCurrentEfficiency,
            lt.LoadTypes.ANY,
            lt.Units.PERCENT,
            output_description="Current efficiency based on the efficiency curve",
        )

        # Total hydrogen production
        self.total_h2_produced: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.SOECTotalHydrogenProduced,
            lt.LoadTypes.HYDROGEN,
            lt.Units.KG,
            output_description="Total hydrogen produced",
        )
        # Total oxygen production
        self.total_o2_produced: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.SOECTotalOxygenProduced,
            lt.LoadTypes.OXYGEN,
            lt.Units.KG,
            output_description="Total oxygen produced",
        )
        # Total water consumed
        self.total_h20_consumed: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.SOECTotalWaterDemand,
            lt.LoadTypes.WATER,
            lt.Units.KG,
            output_description="Total water consumed",
        )
        # Total hydrogen consumed
        self.total_h2_consumed: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.SOFCTotalHydrogenConsumed,
            lt.LoadTypes.HYDROGEN,
            lt.Units.KG,
            output_description="Total hydrogen consumed",
        )
        # Total oxygen consumed
        self.total_o2_consumed: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.SOFCTotalOxygenConsumed,
            lt.LoadTypes.OXYGEN,
            lt.Units.KG,
            output_description="Total oxygen consumed",
        )
        # Total oxygen consumed
        self.total_h20_produced: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.SOFCTotalWaterProduced,
            lt.LoadTypes.WATER,
            lt.Units.KG,
            output_description="Total water produced",
        )
        # Total operating time
        self.total_operating_time_rsco: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.RSOCOperatingTime,
            lt.LoadTypes.TIME,
            lt.Units.HOURS,
            output_description="Total operating time",
        )
        self.current_load_soec: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.SOECCurrentLoad,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            output_description="Current load consumed by SOEC",
        )
        self.total_energy_consumed: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.TotalEnergyConsumed,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KWH,
            output_description="Total load used for hydrogen production",
        )
        self.current_output_sofc: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.SOFCCurrentOutput,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.WATT,
            postprocessing_flag=[
                lt.InandOutputType.ELECTRICITY_PRODUCTION,
                lt.ComponentType.FUEL_CELL,
            ],
            output_description="Current power output SOFC",
        )

        self.total_energy_produced: ComponentOutput = self.add_output(
            self.rsocconfig.name,
            Rsoc.TotalEnergyProduced,
            lt.LoadTypes.ELECTRICITY,
            lt.Units.KWH,
            output_description="Total energy produced for demand",
        )
        # =================================================================================================================================
        # Transfer and storage of states
        self.current_state_soec = 0.0
        self.total_ramp_up_count_state_soec = 0.0
        self.total_ramp_down_count_state_soec = 0.0
        self.total_warm_start_count_soec = 0.0
        self.total_cold_start_count_soec = 0.0
        self.total_warm_start_cycles_soec = 0
        self.total_cold_start_cycles_soec = 0
        self.current_warm_start_count_soec = 0.0
        self.current_cold_start_count_soec = 0.0
        self.total_hydrogen_produced_soec = 0.0
        self.total_oxygen_produced_soec = 0.0
        self.total_water_demand_soec = 0.0
        self.total_energy_soec = 0.0

        self.current_state_sofc = 0.0
        self.total_ramp_up_count_state_sofc = 0.0
        self.total_ramp_down_count_state_sofc = 0.0
        self.total_warm_start_count_sofc = 0.0
        self.total_cold_start_count_sofc = 0.0
        self.total_warm_start_cycles_sofc = 0
        self.total_cold_start_cycles_sofc = 0
        self.current_warm_start_count_sofc = 0.0
        self.current_cold_start_count_sofc = 0.0
        self.total_hydrogen_consumed_sofc = 0.0
        self.total_oxygen_consumed_sofc = 0.0
        self.total_water_produced_sofc = 0.0
        self.total_energy_sofc = 0.0

        self.total_operating_time = 0.0

        self.current_state_soec_previous = self.current_state_soec
        self.total_ramp_up_count_state_soec_previous = self.total_ramp_up_count_state_soec
        self.total_ramp_down_count_state_soec_previous = self.total_ramp_down_count_state_soec
        self.total_warm_start_count_soec_previous = self.total_warm_start_count_soec
        self.total_cold_start_count_soec_previous = self.total_cold_start_count_soec
        self.total_warm_start_cycles_soec_previous = self.total_warm_start_cycles_soec
        self.total_cold_start_cycles_soec_previous = self.total_cold_start_cycles_soec
        self.current_warm_start_count_soec_previous = self.current_warm_start_count_soec
        self.current_cold_start_count_soec_previous = self.current_cold_start_count_soec
        self.total_hydrogen_produced_soec_previous = self.total_hydrogen_produced_soec
        self.total_oxygen_produced_soec_previous = self.total_oxygen_produced_soec
        self.total_water_demand_soec_previous = self.total_water_demand_soec
        self.total_energy_soec_previous = self.total_energy_soec

        self.current_state_sofc_previous = self.current_state_sofc
        self.total_ramp_up_count_state_sofc_previous = self.total_ramp_up_count_state_sofc
        self.total_ramp_down_count_state_sofc_previous = self.total_ramp_down_count_state_sofc
        self.total_warm_start_count_sofc_previous = self.total_warm_start_count_sofc
        self.total_cold_start_count_sofc_previous = self.total_cold_start_count_sofc
        self.total_warm_start_cycles_sofc_previous = self.total_warm_start_cycles_sofc
        self.total_cold_start_cycles_sofc_previous = self.total_cold_start_cycles_sofc
        self.current_warm_start_count_sofc_previous = self.current_warm_start_count_sofc
        self.current_cold_start_count_sofc_previous = self.current_cold_start_count_sofc
        self.total_hydrogen_consumed_sofc_previous = self.total_hydrogen_consumed_sofc
        self.total_oxygen_consumed_sofc_previous = self.total_oxygen_consumed_sofc
        self.total_water_produced_sofc_previous = self.total_water_produced_sofc
        self.total_energy_sofc_previous = self.total_energy_sofc

        self.total_operating_time_previous = self.total_operating_time

    def soec_efficiency(self, name, current_load, min_load, max_load):
        """Efficiency curve data is provided corresponding to the used rSOC system."""
        # Load data from the JSON file
        data_file = os.path.join(utils.HISIMPATH["inputs"], "rSOC_efficiency_curve_data.json")
        with open(data_file, "r", encoding="utf-8") as file:
            data = json.load(file)

        # Check if the provided technology is valid
        if name not in data:
            raise ValueError(f"Invalid rSOC name. Supported names are: {', '.join(data.keys())}")

        # Extract the x and y data points for the selected technology
        load_percentage = data[name]["load_percentage_soec"]
        sys_eff = data[name]["sys_eff_soec"]

        if min_load <= current_load <= max_load:
            current_load_percentage = current_load / max_load
            # Interpolation
            current_sys_eff_soec = float(np.interp(current_load_percentage, load_percentage, sys_eff))

        else:
            current_sys_eff_soec = 0.0

        return current_sys_eff_soec

    def sofc_efficiency(self, name, current_demand, min_power, max_power):
        """Efficiency curve data is provided corresponding to the used rSOC system."""
        # Load data from the JSON file
        data_file = os.path.join(utils.HISIMPATH["inputs"], "rSOC_efficiency_curve_data.json")
        with open(data_file, "r", encoding="utf-8") as file:
            data = json.load(file)

        # Check if the provided technology is valid
        if name not in data:
            raise ValueError(f"Invalid rSOC name. Supported names are: {', '.join(data.keys())}")

        # Extract the x and y data points for the selected technology
        load_percentage = data[name]["load_percentage_sofc"]
        sys_eff = data[name]["sys_eff_sofc"]

        if min_power < current_demand <= max_power:
            current_demand_percentage = current_demand / max_power
            # Interpolation
            current_sys_eff_sofc = float(np.interp(current_demand_percentage, load_percentage, sys_eff))

        else:
            current_sys_eff_sofc = 0.0

        return current_sys_eff_sofc

    def h2_production_rate(self, current_sys_eff_soec, current_load):
        """Based on the calculated efficiency the current h2 production rate is calculated."""
        lhv_h2 = 33.33  # [kWh/kg]

        h2_production_rate = (current_sys_eff_soec * current_load) / (lhv_h2 * 3600)  # [kg/s]
        return h2_production_rate

    def h2_consumption_rate(self, current_sys_eff_sofc, current_demand):
        """Based on the calculated efficiency the current h2 consumption rate is calculated."""
        lhv_h2 = 33.33  # [kWh/kg]
        if current_sys_eff_sofc > 0.0 and current_demand >= self.min_power_sofc:
            h2_consumption_rate = current_demand / (current_sys_eff_sofc * lhv_h2 * 3600)  # [kg/s]
        else:
            h2_consumption_rate = 0.0
        return h2_consumption_rate

    def oxygen_rate(self, current_h2_rate):
        """Returns the mass flow rate of oxygen, based on the current hydrogen flow rate."""
        m_o2 = 31.9988  # g/mol
        m_h2 = 2.01588  # g/mol
        o2_flow_rate = (
            (m_o2 / m_h2) * 0.5 * current_h2_rate
        )  # Wang (2021) - Thermodynamic analysis of solid oxide electrolyzer integration with engine waste heat recovery for hydrogen production
        return o2_flow_rate

    def water_rate(self, current_h2_rate):
        """Returns the water mass flow rate, based on the current hydrogen flow rate."""
        m_h2o = 18.01528  # g/mol
        m_h2 = 2.01588  # g/mol
        h2o_flow_rate = (
            m_h2o / m_h2
        ) * current_h2_rate  # Wang (2021) - Thermodynamic analysis of solid oxide electrolyzer integration with engine waste heat recovery for hydrogen production
        return h2o_flow_rate

    def i_save_state(self) -> None:
        """Saves the current state."""
        self.current_state_soec_previous = self.current_state_soec
        self.total_ramp_up_count_state_soec_previous = self.total_ramp_up_count_state_soec
        self.total_ramp_down_count_state_soec_previous = self.total_ramp_down_count_state_soec
        self.total_warm_start_count_soec_previous = self.total_warm_start_count_soec
        self.total_cold_start_count_soec_previous = self.total_cold_start_count_soec
        self.total_warm_start_cycles_soec_previous = self.total_warm_start_cycles_soec
        self.total_cold_start_cycles_soec_previous = self.total_cold_start_cycles_soec
        self.current_warm_start_count_soec_previous = self.current_warm_start_count_soec
        self.current_cold_start_count_soec_previous = self.current_cold_start_count_soec
        self.total_hydrogen_produced_soec_previous = self.total_hydrogen_produced_soec
        self.total_oxygen_produced_soec_previous = self.total_oxygen_produced_soec
        self.total_water_demand_soec_previous = self.total_water_demand_soec
        self.total_energy_soec_previous = self.total_energy_soec

        self.current_state_sofc_previous = self.current_state_sofc
        self.total_ramp_up_count_state_sofc_previous = self.total_ramp_up_count_state_sofc
        self.total_ramp_down_count_state_sofc_previous = self.total_ramp_down_count_state_sofc
        self.total_warm_start_count_sofc_previous = self.total_warm_start_count_sofc
        self.total_cold_start_count_sofc_previous = self.total_cold_start_count_sofc
        self.total_warm_start_cycles_sofc_previous = self.total_warm_start_cycles_sofc
        self.total_cold_start_cycles_sofc_previous = self.total_cold_start_cycles_sofc
        self.current_warm_start_count_sofc_previous = self.current_warm_start_count_sofc
        self.current_cold_start_count_sofc_previous = self.current_cold_start_count_sofc
        self.total_hydrogen_consumed_sofc_previous = self.total_hydrogen_consumed_sofc
        self.total_oxygen_consumed_sofc_previous = self.total_oxygen_consumed_sofc
        self.total_water_produced_sofc_previous = self.total_water_produced_sofc
        self.total_energy_sofc_previous = self.total_energy_sofc

        self.total_operating_time_previous = self.total_operating_time

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_restore_state(self) -> None:
        """Restores previous state."""
        self.current_state_soec = self.current_state_soec_previous
        self.total_ramp_up_count_state_soec = self.total_ramp_up_count_state_soec_previous
        self.total_ramp_down_count_state_soec = self.total_ramp_down_count_state_soec_previous
        self.total_warm_start_count_soec = self.total_warm_start_count_soec_previous
        self.total_cold_start_count_soec = self.total_cold_start_count_soec_previous
        self.total_warm_start_cycles_soec = self.total_warm_start_cycles_soec_previous
        self.total_cold_start_cycles_soec = self.total_cold_start_cycles_soec_previous
        self.current_warm_start_count_soec = self.current_warm_start_count_soec_previous
        self.current_cold_start_count_soec = self.current_cold_start_count_soec_previous
        self.total_hydrogen_produced_soec = self.total_hydrogen_produced_soec_previous
        self.total_oxygen_produced_soec = self.total_oxygen_produced_soec_previous
        self.total_water_demand_soec = self.total_water_demand_soec_previous
        self.total_energy_soec = self.total_energy_soec_previous

        self.current_state_sofc = self.current_state_sofc_previous
        self.total_ramp_up_count_state_sofc = self.total_ramp_up_count_state_sofc_previous
        self.total_ramp_down_count_state_sofc = self.total_ramp_down_count_state_sofc_previous
        self.total_warm_start_count_sofc = self.total_warm_start_count_sofc_previous
        self.total_cold_start_count_sofc = self.total_cold_start_count_sofc_previous
        self.total_warm_start_cycles_sofc = self.total_warm_start_cycles_sofc_previous
        self.total_cold_start_cycles_sofc = self.total_cold_start_cycles_sofc_previous
        self.current_warm_start_count_sofc = self.current_warm_start_count_sofc_previous
        self.current_cold_start_count_sofc = self.current_cold_start_count_sofc_previous
        self.total_hydrogen_consumed_sofc = self.total_hydrogen_consumed_sofc_previous
        self.total_oxygen_consumed_sofc = self.total_oxygen_consumed_sofc_previous
        self.total_water_produced_sofc = self.total_water_produced_sofc_previous
        self.total_energy_sofc = self.total_energy_sofc_previous

        self.total_operating_time = self.total_operating_time_previous

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulate the component."""

        seconds_per_timestep = self.my_simulation_parameters.seconds_per_timestep  # [s/timestep]

        if stsv.get_input_value(self.power_input) < 0.0:
            power_to_soec = abs(stsv.get_input_value(self.power_input))
            demand_to_sofc = 0.0
        elif stsv.get_input_value(self.power_input) > 0.0:
            demand_to_sofc = abs(stsv.get_input_value(self.power_input))
            power_to_soec = 0.0
        else:
            power_to_soec = 0.0
            demand_to_sofc = 0.0
        rsoc_state = stsv.get_input_value(self.input_state_rsoc)

        if rsoc_state == 1:
            self.total_operating_time += self.my_simulation_parameters.seconds_per_timestep / 3600
            stsv.set_output_value(self.total_operating_time_rsco, self.total_operating_time)
        else:
            self.total_operating_time += 0.0
            stsv.set_output_value(self.total_operating_time_rsco, self.total_operating_time)

        if power_to_soec != 0.0:
            # SOEC operation
            # Variable
            self.current_state_sofc = 0.0
            name = self.name
            nominal_load = self.nom_load_soec
            min_load = self.min_load_soec
            ramp_up_rate = self.ramp_up_rate_soec
            ramp_down_rate = self.ramp_down_rate_soec

            # ramp up per timestep calculation
            ramp_up_per_timestep = nominal_load * ramp_up_rate * seconds_per_timestep
            total_ramp_up_per_timestep = ramp_up_per_timestep

            # ramp down per timestep calculation
            ramp_down_per_timestep = nominal_load * ramp_down_rate * seconds_per_timestep
            total_ramp_down_per_timestep = ramp_down_per_timestep

            # calculating load input
            new_load = abs(power_to_soec - self.current_state_soec)

            current_sys_eff_soec = 0.0
            h2_production_rate = 0.0

            if rsoc_state == 1:
                """the ramping process"""
                if new_load == nominal_load:
                    self.current_state_soec = nominal_load
                    self.total_ramp_up_count_state_soec += 0
                    self.total_ramp_down_count_state_soec += 0
                elif new_load == 0:
                    self.current_state_soec = self.current_state_soec
                    self.total_ramp_up_count_state_soec += 0
                    self.total_ramp_down_count_state_soec += 0

                # Ramping up
                if new_load >= total_ramp_up_per_timestep and self.current_state_soec < power_to_soec:
                    self.total_ramp_up_count_state_soec += seconds_per_timestep
                    self.current_state_soec += total_ramp_up_per_timestep

                elif self.current_state_soec < power_to_soec and new_load < total_ramp_up_per_timestep:
                    percentage_ramp_up_per_timestep = new_load / total_ramp_up_per_timestep
                    self.total_ramp_up_count_state_soec += percentage_ramp_up_per_timestep * seconds_per_timestep

                    if self.current_state_soec == 0:
                        self.current_state_soec += power_to_soec

                    else:
                        self.current_state_soec += new_load

                # Ramping down
                elif total_ramp_down_per_timestep <= new_load and power_to_soec < self.current_state_soec:
                    self.total_ramp_down_count_state_soec += seconds_per_timestep
                    self.current_state_soec -= new_load

                elif power_to_soec < self.current_state_soec and new_load < total_ramp_down_per_timestep:
                    percentage_ramp_down_per_timestep = new_load / total_ramp_down_per_timestep
                    self.total_ramp_down_count_state_soec += percentage_ramp_down_per_timestep * seconds_per_timestep
                    self.current_state_soec -= new_load

                current_sys_eff_soec = self.soec_efficiency(name, self.current_state_soec, min_load, self.max_load_soec)
                h2_production_rate = self.h2_production_rate(current_sys_eff_soec, self.current_state_soec)

            elif rsoc_state == 0:
                self.total_ramp_up_count_state_soec += 0
                self.total_ramp_down_count_state_soec += 0
                self.current_state_soec = 2.315

            elif rsoc_state == -1:
                self.total_ramp_up_count_state_soec += 0
                self.total_ramp_down_count_state_soec += 0
                self.current_state_soec = 0.0

            else:
                raise ValueError("`rsoc_state` needs to be one of [-1, 0, 1]")

            # current_sys_eff_soec = self.soec_efficiency(name, self.current_state_soec, min_load, self.max_load_soec)
            # h2_production_rate = self.h2_production_rate(current_sys_eff_soec, self.current_state_soec)
            o2_flow_rate_soec = self.oxygen_rate(h2_production_rate)
            h2o_flow_rate_soec = self.water_rate(h2_production_rate)

            self.total_hydrogen_produced_soec += h2_production_rate * seconds_per_timestep
            self.total_oxygen_produced_soec += o2_flow_rate_soec * seconds_per_timestep
            self.total_water_demand_soec += h2o_flow_rate_soec * seconds_per_timestep

            stsv.set_output_value(self.soec_hydrogen_flow_rate, h2_production_rate)
            stsv.set_output_value(self.soec_current_efficiency_state, current_sys_eff_soec)

            # self.total_hydrogen_consumed_sofc += 0.0
            self.total_oxygen_consumed_sofc += 0.0
            self.total_water_produced_sofc += 0.0

            h2_consumption_rate = 0.0
            o2_flow_rate_sofc = 0.0
            h2o_flow_rate_sofc = 0.0

        elif demand_to_sofc != 0:
            # SOFC implementation start
            # Variable
            self.current_state_soec = 0.0

            name = self.name

            seconds_per_timestep = self.my_simulation_parameters.seconds_per_timestep  # [s/timestep]

            # ramp up per timestep calculation
            ramp_up_per_timestep = self.nom_power_sofc * self.ramp_up_rate_sofc * seconds_per_timestep
            total_ramp_up_per_timestep = ramp_up_per_timestep

            # ramp down per timestep calculation
            ramp_down_per_timestep = self.nom_power_sofc * self.ramp_down_rate_sofc * seconds_per_timestep
            total_ramp_down_per_timestep = ramp_down_per_timestep

            # calculating load input
            new_load = abs(demand_to_sofc - self.current_state_sofc)

            if rsoc_state == 1:
                """the ramping process"""
                if new_load == self.nom_power_sofc:
                    self.current_state_sofc = self.nom_power_sofc
                    self.total_ramp_up_count_state_sofc += 0
                    self.total_ramp_down_count_state_sofc += 0

                elif new_load == 0:
                    self.current_state_sofc = self.current_state_sofc
                    self.total_ramp_up_count_state_sofc += 0
                    self.total_ramp_down_count_state_sofc += 0

                # Ramping up
                if new_load >= total_ramp_up_per_timestep and self.current_state_sofc < demand_to_sofc:
                    self.total_ramp_up_count_state_sofc += seconds_per_timestep
                    self.current_state_sofc += total_ramp_up_per_timestep

                elif self.current_state_sofc < demand_to_sofc and new_load < total_ramp_up_per_timestep:
                    percentage_ramp_up_per_timestep = new_load / total_ramp_up_per_timestep
                    self.total_ramp_up_count_state_sofc += percentage_ramp_up_per_timestep * seconds_per_timestep

                    if self.current_state_sofc == 0:
                        self.current_state_sofc += demand_to_sofc

                    else:
                        self.current_state_sofc += new_load

                # Ramping down
                elif total_ramp_down_per_timestep <= new_load and demand_to_sofc < self.current_state_sofc:
                    self.total_ramp_down_count_state_sofc += seconds_per_timestep
                    self.current_state_sofc -= new_load

                elif demand_to_sofc < self.current_state_sofc and new_load < total_ramp_down_per_timestep:
                    percentage_ramp_down_per_timestep = new_load / total_ramp_down_per_timestep
                    self.total_ramp_down_count_state_sofc += percentage_ramp_down_per_timestep * seconds_per_timestep
                    self.current_state_sofc -= new_load

            elif rsoc_state == 0:
                self.total_ramp_up_count_state_sofc += 0
                self.total_ramp_down_count_state_sofc += 0
                self.current_state_sofc = 0.0
                h2_consumption_rate = 0.0

            elif rsoc_state == -1:
                self.total_ramp_up_count_state_sofc += 0
                self.total_ramp_down_count_state_sofc += 0
                self.current_state_sofc = 0.0
                h2_consumption_rate = 0.0

            current_sys_eff_sofc = self.sofc_efficiency(
                name, self.current_state_sofc, self.min_power_sofc, self.max_power_sofc
            )
            h2_consumption_rate = self.h2_consumption_rate(current_sys_eff_sofc, self.current_state_sofc)
            o2_flow_rate_sofc = self.oxygen_rate(h2_consumption_rate)
            h2o_flow_rate_sofc = self.water_rate(h2_consumption_rate)

            # self.total_hydrogen_consumed_sofc += h2_consumption_rate * seconds_per_timestep
            self.total_oxygen_consumed_sofc += o2_flow_rate_sofc * seconds_per_timestep
            self.total_water_produced_sofc += h2o_flow_rate_sofc * seconds_per_timestep

            stsv.set_output_value(self.sofc_hydrogen_flow_rate, h2_consumption_rate)
            stsv.set_output_value(self.sofc_current_efficiency_state, current_sys_eff_sofc)

            # end

            self.total_hydrogen_produced_soec += 0.0
            h2_production_rate = 0.0
            o2_flow_rate_soec = 0.0
            h2o_flow_rate_soec = 0.0
            self.total_oxygen_produced_soec += 0.0
            self.total_water_demand_soec += 0.0

        else:
            stsv.set_output_value(self.sofc_hydrogen_flow_rate, 0.0)
            stsv.set_output_value(self.sofc_current_efficiency_state, 0.0)
            stsv.set_output_value(self.soec_hydrogen_flow_rate, 0.0)
            stsv.set_output_value(self.soec_current_efficiency_state, 0.0)
            self.current_state_sofc = 0.0
            self.current_state_soec = 2.315
            # stsv.set_output_value(self.current_output_sofc, -2315) #for WATT output in system setup
            # stsv.set_output_value(self.current_load_soec, 2315)

            self.total_hydrogen_produced_soec += 0.0
            # self.total_hydrogen_consumed_sofc += 0.0
            h2_consumption_rate = 0.0
            h2_production_rate = 0.0
            o2_flow_rate_soec = 0.0
            h2o_flow_rate_soec = 0.0
            o2_flow_rate_sofc = 0.0
            h2o_flow_rate_sofc = 0.0

        stsv.set_output_value(self.total_h2_produced, self.total_hydrogen_produced_soec)
        stsv.set_output_value(self.total_o2_produced, self.total_oxygen_produced_soec)
        stsv.set_output_value(self.total_h20_consumed, self.total_water_demand_soec)

        self.total_hydrogen_consumed_sofc += h2_consumption_rate * seconds_per_timestep
        stsv.set_output_value(self.total_h2_consumed, self.total_hydrogen_consumed_sofc)
        stsv.set_output_value(self.total_o2_consumed, self.total_oxygen_consumed_sofc)
        stsv.set_output_value(self.total_h20_produced, self.total_water_produced_sofc)

        stsv.set_output_value(
            self.current_output_sofc, (self.current_state_sofc * 1000)
        )  # for WATT output in system setup
        stsv.set_output_value(
            self.current_load_soec, (self.current_state_soec * 1000)
        )  # for WATT output in system setup

        self.total_energy_soec += self.current_state_soec * (seconds_per_timestep / 3600)
        stsv.set_output_value(self.total_energy_consumed, self.total_energy_soec)
        self.total_energy_sofc += self.current_state_sofc * (seconds_per_timestep / 3600)
        stsv.set_output_value(self.total_energy_produced, self.total_energy_sofc)

    def write_to_report(self):
        """Writes the information of the current component to the report."""
        lines = []
        for config_string in self.rsocconfig.get_string_dict():
            lines.append(config_string)
        lines.append("Component Name" + str(self.component_name))
        lines.append("Total operating time during simulation: " + str(self.total_operating_time) + " [h]")
        lines.append("Total hydrogen produced during simulation: " + str(self.total_hydrogen_produced_soec) + " [kg]")
        lines.append("Total hydrogen consumed during simulation: " + str(self.total_hydrogen_consumed_sofc) + " [kg]")
        lines.append("Total oxygen produced during simulation: " + str(self.total_oxygen_produced_soec) + " [kg]")
        lines.append("Total oxygen consumed during simulation: " + str(self.total_oxygen_consumed_sofc) + " [kg]")
        lines.append("Total water demand during simulation: " + str(self.total_water_demand_soec) + " [kg]")
        lines.append("Total water demand during simulation: " + str(self.total_water_produced_sofc) + " [kg]")
        lines.append("Total energy consumed during simulation: " + str(self.total_energy_soec) + " [kg]")
        lines.append("Total energy produced during simulation: " + str(self.total_energy_sofc) + " [kWh]")
        lines.append("Total ramp-up time during simulation: " + str(self.total_ramp_up_count_state_soec) + " [sec]")
        lines.append("Total ramp-up time during simulation: " + str(self.total_ramp_up_count_state_sofc) + " [sec]")
        lines.append("Total ramp-down time during simulation: " + str(self.total_ramp_down_count_state_soec) + " [sec]")
        lines.append("Total ramp-down time during simulation: " + str(self.total_ramp_down_count_state_sofc) + " [sec]")
        return lines



""" Logging functionality for all of HiSim. """
# clean
from enum import IntEnum
import os

LOGGING_LEVEL = 3
LOGGING_DEFAULT_PATH: str = r"../logs/"


class LogPrio(IntEnum):

    """Define a logging priority."""

    ERROR = 1
    WARNING = 2
    INFORMATION = 3
    DEBUG = 4
    PROFILE = 5
    TRACE = 6


def error(message: str, logging_message_path: str = LOGGING_DEFAULT_PATH) -> None:
    """Log an error message."""
    log(LogPrio.ERROR, message, logging_message_path)


def warning(message: str, logging_message_path: str = LOGGING_DEFAULT_PATH) -> None:
    """Log a warning message."""
    log(LogPrio.WARNING, message, logging_message_path)


def information(message: str, logging_message_path: str = LOGGING_DEFAULT_PATH) -> None:
    """Log a information message."""
    log(LogPrio.INFORMATION, message, logging_message_path)


def trace(message: str, logging_message_path: str = LOGGING_DEFAULT_PATH) -> None:
    """Log a trace message."""
    log(LogPrio.TRACE, message, logging_message_path)


def debug(message: str, logging_message_path: str = LOGGING_DEFAULT_PATH) -> None:
    """Log a debug message."""
    log(LogPrio.DEBUG, message, logging_message_path)


def profile(message: str, logging_message_path: str = LOGGING_DEFAULT_PATH) -> None:
    """Log a profile message."""
    log(LogPrio.PROFILE, message, logging_message_path)
    log_profile_file(message, logging_message_path)


def log(prio: int, message: str, logging_message_path: str = LOGGING_DEFAULT_PATH) -> None:
    """Write and print a log message."""
    # if(prio < LogPrio.Debug):
    prio_string: str
    if prio == LogPrio.ERROR:
        prio_string = "ERR"
    elif prio == LogPrio.WARNING:
        prio_string = "WRN"
    elif prio == LogPrio.INFORMATION:
        prio_string = "IFO"
    elif prio == LogPrio.DEBUG:
        prio_string = "DBG"
    elif prio == LogPrio.PROFILE:
        prio_string = "PRF"
    elif prio == LogPrio.TRACE:
        prio_string = "TRC"
    else:
        raise ValueError("Unknown log priority: " + str(prio))
    if prio <= LOGGING_LEVEL:
        print(str(prio_string) + ":" + message)

    if not os.path.exists(logging_message_path):
        os.makedirs(logging_message_path)

    file_name = os.path.join(logging_message_path, "hisim_simulation.log")
    try:
        with open(file_name, "a", encoding="utf-8") as filestream:
            filestream.write(message + "\n")
    except Exception:
        print("hisim_simulation.log could not be appended. "
              "This might happen when too many simultaneous simulations are running.")


def log_profile_file(message: str, logging_message_path: str = LOGGING_DEFAULT_PATH) -> None:
    """Write log message to logfile."""

    if not os.path.exists(logging_message_path):
        os.makedirs(logging_message_path)

    file_name = os.path.join(logging_message_path, "profiling_timeuse.log")
    try:
        with open(file_name, "a", encoding="utf-8") as filestream:
            filestream.write(message + "\n")
    except Exception:
        print("profiling_timeuse.log could not be appended. "
              "This might happen when too many simultaneous simulations are running.")



"""Controller l1 heat old module."""

# clean

# Generic/Built-in
from typing import Any
from dataclasses import dataclass
from dataclasses_json import dataclass_json

from hisim import component as cp
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters
from hisim import utils


@dataclass_json
@dataclass
class ControllerHeatConfig(cp.ConfigBase):
    """Configuration of the Controller Heat class."""

    @classmethod
    def get_main_classname(cls):
        """Return the full class name of the base class."""
        return ControllerHeat.get_full_classname()

    building_name: str
    name: str
    temperature_storage_target_warm_water: float
    temperature_storage_target_heating_water: float
    temperature_storage_target_hysteresis_ww: float
    temperature_storage_target_hysteresis_hw: float

    @classmethod
    def get_default_controller_heat_l1(
        cls,
        building_name: str = "BUI1",
    ) -> Any:
        """Get a default Building."""
        config = ControllerHeatConfig(
            building_name=building_name,
            name="ControllerHeatL1",
            temperature_storage_target_warm_water=50,
            temperature_storage_target_heating_water=35,
            temperature_storage_target_hysteresis_ww=45,
            temperature_storage_target_hysteresis_hw=35,
        )
        return config


class ControllerState:
    """Controller state.

    Save State if Heater Components were supposed to run
    in last timestep (Control_Signal).
    Saves timestep of hysteresis of storages and  the
    changing target temperature of storages. The
    target temperature in states changes when target
    temperature of storage is reached.
    """

    def __init__(
        self,
        control_signal_gas_heater: float,
        control_signal_chp: float,
        control_signal_heat_pump: float,
        temperature_storage_target_ww_c: float,
        temperature_storage_target_hw_c: float,
        timestep_of_hysteresis_ww: int,
        timestep_of_hysteresis_hw: int,
    ) -> None:
        """Initialize the class."""
        self.control_signal_gas_heater: float = control_signal_gas_heater
        self.control_signal_chp: float = control_signal_chp
        self.control_signal_heat_pump: float = control_signal_heat_pump
        self.temperature_storage_target_ww_c: float = temperature_storage_target_ww_c
        self.temperature_storage_target_hw_c: float = temperature_storage_target_hw_c
        self.timestep_of_hysteresis_ww: int = timestep_of_hysteresis_ww
        self.timestep_of_hysteresis_hw: int = timestep_of_hysteresis_hw

    def clone(self) -> "ControllerState":
        """Clones the state."""
        return ControllerState(
            control_signal_gas_heater=self.control_signal_gas_heater,
            control_signal_chp=self.control_signal_chp,
            control_signal_heat_pump=self.control_signal_heat_pump,
            temperature_storage_target_ww_c=self.temperature_storage_target_ww_c,
            temperature_storage_target_hw_c=self.temperature_storage_target_hw_c,
            timestep_of_hysteresis_ww=self.timestep_of_hysteresis_ww,
            timestep_of_hysteresis_hw=self.timestep_of_hysteresis_hw,
        )


class ControllerHeat(cp.Component):
    """Controller heat.

    Controlls energy flows for heat demand.
    Heat Demand will be controlled by the storage temperature.
    For this storage provides heat for a load profile or for the
    building. As well Heat Demand can be simulated without storage.
    """

    # Inputs
    StorageTemperatureHeatingWater = "StorageTemperatureHeatingWater"
    StorageTemperatureWarmWater = "StorageTemperatureWarmWater"
    ResidenceTemperature = "ResidenceTemperature"
    ThermalDemandBuilding = "ThermalDemandBuilding"

    # Outputs
    ControlSignalGasHeater = "ControlSignalGasHeater"
    ControlSignalChp = "ControlSignalChp"
    ControlSignalHeatPump = "ControlSignalHeatPump"
    ControlSignalChooseStorage = "ControlSignalChooseStorage"

    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: ControllerHeatConfig,
        my_display_config: cp.DisplayConfig = cp.DisplayConfig(),
    ) -> None:
        """Initialize the class."""
        self.controller_heat_config = config
        self.my_simulation_parameters = my_simulation_parameters
        self.config = config
        component_name = self.get_component_name()
        super().__init__(
            name=component_name,
            my_simulation_parameters=my_simulation_parameters,
            my_config=config,
            my_display_config=my_display_config,
        )
        self.mode: Any
        self.temperature_storage_target_warm_water = self.controller_heat_config.temperature_storage_target_warm_water
        self.temperature_storage_target_heating_water = (
            self.controller_heat_config.temperature_storage_target_heating_water
        )
        self.temperature_storage_target_hysteresis_hw = (
            self.controller_heat_config.temperature_storage_target_hysteresis_hw
        )
        self.temperature_storage_target_hysteresis_ww = (
            self.controller_heat_config.temperature_storage_target_hysteresis_ww
        )

        self.state = ControllerState(
            control_signal_heat_pump=0,
            control_signal_gas_heater=0,
            control_signal_chp=0,
            temperature_storage_target_ww_c=self.temperature_storage_target_warm_water,
            temperature_storage_target_hw_c=self.temperature_storage_target_heating_water,
            timestep_of_hysteresis_ww=0,
            timestep_of_hysteresis_hw=0,
        )
        self.previous_state = self.state.clone()

        # Inputs
        self.temperature_storage_warm_water_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.StorageTemperatureWarmWater,
            lt.LoadTypes.WATER,
            lt.Units.CELSIUS,
            False,
        )
        self.temperature_storage_heating_water_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.StorageTemperatureHeatingWater,
            lt.LoadTypes.WATER,
            lt.Units.CELSIUS,
            False,
        )
        self.temperature_residence_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ResidenceTemperature,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            False,
        )

        # Outputs
        self.control_signal_gas_heater_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ControlSignalGasHeater,
            load_type=lt.LoadTypes.ANY,
            unit=lt.Units.PERCENT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ControlSignalGasHeater} will follow.",
        )
        self.control_signal_chp_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ControlSignalChp,
            load_type=lt.LoadTypes.ANY,
            unit=lt.Units.PERCENT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ControlSignalChp} will follow.",
        )
        self.control_signal_heat_pump_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ControlSignalHeatPump,
            load_type=lt.LoadTypes.ANY,
            unit=lt.Units.PERCENT,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ControlSignalHeatPump} will follow.",
        )
        self.control_signal_choose_storage_channel: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ControlSignalChooseStorage,
            load_type=lt.LoadTypes.ANY,
            unit=lt.Units.ANY,
            sankey_flow_direction=False,
            output_description=f"here a description for {self.ControlSignalChooseStorage} will follow.",
        )

    def build(self, mode: Any) -> None:
        """Build function."""
        self.mode = mode

    def write_to_report(self):
        """Write to report."""
        return self.controller_heat_config.get_string_dict()

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_save_state(self) -> None:
        """Saves the state."""
        self.previous_state = self.state.clone()

    def i_restore_state(self) -> None:
        """Restores the state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    # Simulates and defines the control signals to heat up storages
    # work as a 2-point Ruler with Hysteresis
    def simulate_storage(
        self,
        delta_temperature: float,
        stsv: cp.SingleTimeStepValues,
        timestep: int,
        temperature_storage: float,
        temperature_storage_target: float,
        temperature_storage_target_hysteresis: float,
        temperature_storage_target_c: float,
        timestep_of_hysteresis: int,
    ) -> Any:
        """Simulates the storage."""
        control_signal_chp: float = 0
        control_signal_gas_heater: float = 0
        control_signal_heat_pump: float = 0

        max_temperature_limit = 5
        if temperature_storage > 0:
            if delta_temperature > max_temperature_limit:
                control_signal_heat_pump = 1
                control_signal_chp = 1
                control_signal_gas_heater = 1
                temperature_storage_target_c = temperature_storage_target

            elif 0 < delta_temperature <= max_temperature_limit:
                control_signal_heat_pump = 1
                control_signal_chp = 1
                control_signal_gas_heater = 1

                if self.state.control_signal_chp < control_signal_chp:
                    control_signal_chp = 1
                elif self.state.control_signal_gas_heater < control_signal_gas_heater:
                    control_signal_gas_heater = 1
                temperature_storage_target_c = temperature_storage_target

                # Storage warm enough. Try to turn off Heaters
            elif delta_temperature <= 0:
                if temperature_storage_target_c == temperature_storage_target and timestep_of_hysteresis != timestep:
                    temperature_storage_target_c = temperature_storage_target_hysteresis
                    timestep_of_hysteresis = timestep
                elif temperature_storage_target_c != temperature_storage_target and timestep_of_hysteresis != timestep:
                    control_signal_heat_pump = 0
                    control_signal_gas_heater = 0
                    control_signal_chp = 0

        self.state.control_signal_gas_heater = control_signal_gas_heater
        self.state.control_signal_chp = control_signal_chp
        self.state.control_signal_heat_pump = control_signal_heat_pump
        stsv.set_output_value(self.control_signal_heat_pump_channel, control_signal_heat_pump)
        stsv.set_output_value(self.control_signal_gas_heater_channel, control_signal_gas_heater)
        stsv.set_output_value(self.control_signal_chp_channel, control_signal_chp)

        return temperature_storage_target_c, timestep_of_hysteresis

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        if force_convergence:
            return
        # Heat
        # Logic of regulating HeatDemand:
        # First heat up WarmWaterStorage->more important, than heat up HeatingWater
        # But only one Storage can be heated up in a TimeStep!
        # Simulate WarmWater
        delta_temperature_ww = self.state.temperature_storage_target_ww_c - stsv.get_input_value(
            self.temperature_storage_warm_water_channel
        )
        delta_temperature_hw = self.state.temperature_storage_target_hw_c - stsv.get_input_value(
            self.temperature_storage_heating_water_channel
        )
        if (
            stsv.get_input_value(self.temperature_storage_warm_water_channel) == 0
            and stsv.get_input_value(self.temperature_storage_heating_water_channel) != 0
        ):
            control_signal_choose_storage = 2
        elif (
            stsv.get_input_value(self.temperature_storage_warm_water_channel) != 0
            and stsv.get_input_value(self.temperature_storage_heating_water_channel) == 0
        ):
            control_signal_choose_storage = 1
        else:
            # Choose which Storage should be heated up
            if delta_temperature_ww >= 0 and delta_temperature_hw >= 0:
                if delta_temperature_hw <= delta_temperature_ww:
                    control_signal_choose_storage = 1
                else:
                    control_signal_choose_storage = 2
            elif delta_temperature_ww < 0 and delta_temperature_hw < 0:
                if delta_temperature_hw <= delta_temperature_ww:
                    control_signal_choose_storage = 1
                else:
                    control_signal_choose_storage = 2
            elif delta_temperature_ww <= 0 <= delta_temperature_hw:
                control_signal_choose_storage = 2
            elif delta_temperature_hw <= 0 <= delta_temperature_ww:
                control_signal_choose_storage = 1

        # Heats up storage
        if control_signal_choose_storage == 1:
            (
                self.state.temperature_storage_target_ww_c,
                self.state.timestep_of_hysteresis_ww,
            ) = self.simulate_storage(
                stsv=stsv,
                delta_temperature=delta_temperature_ww,
                timestep=timestep,
                temperature_storage=stsv.get_input_value(self.temperature_storage_warm_water_channel),
                temperature_storage_target=self.temperature_storage_target_warm_water,
                temperature_storage_target_hysteresis=self.temperature_storage_target_hysteresis_ww,
                temperature_storage_target_c=self.state.temperature_storage_target_ww_c,
                timestep_of_hysteresis=self.state.timestep_of_hysteresis_ww,
            )
        elif control_signal_choose_storage == 2:
            delta_temperature_hw = self.state.temperature_storage_target_hw_c - stsv.get_input_value(
                self.temperature_storage_heating_water_channel
            )
            (
                self.state.temperature_storage_target_hw_c,
                self.state.timestep_of_hysteresis_hw,
            ) = self.simulate_storage(
                stsv=stsv,
                delta_temperature=delta_temperature_hw,
                timestep=timestep,
                temperature_storage=stsv.get_input_value(self.temperature_storage_heating_water_channel),
                temperature_storage_target=self.temperature_storage_target_heating_water,
                temperature_storage_target_hysteresis=self.temperature_storage_target_hysteresis_hw,
                temperature_storage_target_c=self.state.temperature_storage_target_hw_c,
                timestep_of_hysteresis=self.state.timestep_of_hysteresis_hw,
            )

        stsv.set_output_value(self.control_signal_choose_storage_channel, control_signal_choose_storage)



# clean

"""Classes to provide the structure for the KPi generation."""
from typing import Optional, Union, List, Tuple
from enum import Enum
from dataclasses import dataclass
from dataclass_wizard import JSONWizard
import pandas as pd
import numpy as np


class KpiTagEnumClass(Enum):
    """Determine KPI tags as enums."""

    GENERAL = "General"
    COSTS = "Costs"
    EMISSIONS = "Emissions"
    BUILDING = "Building"
    BATTERY = "Battery"
    HEAT_DISTRIBUTION_SYSTEM = "Heat Distribution System"
    HEATPUMP_SPACE_HEATING = "Heat Pump For Space Heating"
    HEATPUMP_DOMESTIC_HOT_WATER = "Heat Pump For Domestic Hot Water"
    HEATPUMP_SPACE_HEATING_AND_DOMESTIC_HOT_WATER = "Heat Pump For SH and DHW"
    RESIDENTS = "Residents"
    GAS_HEATER_SPACE_HEATING = "Gas Heater For Space Heating"
    GAS_HEATER_DOMESTIC_HOT_WATER = "Gas Heater For Domestic Hot Water"
    GAS_METER = "Gas Meter"
    HEATING_METER = "Heating Meter"
    ELECTRICITY_METER = "Electricity Meter"
    CAR = "Car"
    CAR_BATTERY = "Car Battery"
    ROOFTOP_PV = "Rooftop PV"
    STORAGE_DOMESTIC_HOT_WATER = "Storage For Domestic Hot Water"
    STORAGE_HOT_WATER_SPACE_HEATING = "Storage For Space Heating Hot Water"
    WINDTURBINE = "Wind Turbine"
    SMART_DEVICE = "Smart Device"
    EMS = "Energy Management System"
    ELECTRICITY_GRID = "Electricity Grid"
    THERMAL_GRID = "Thermal Grid"
    COSTS_DISTRICT_GRID = "Costs Of District Grid"
    EMISSIONS_DISTRICT_GRID = "Emissions Of District Grid"
    CONTRACTING = "Contracting"
    GENERIC_HEAT_SOURCE = "Generic Heat Source"


@dataclass
class KpiEntry(JSONWizard):
    """Class for storing one kpi entry."""

    name: str
    unit: str
    value: Optional[float]
    description: Optional[str] = None
    tag: Optional[KpiTagEnumClass] = None


class KpiHelperClass:
    """Class for providing some helper fucntions for calculating KPIs."""

    @staticmethod
    def compute_total_energy_from_power_timeseries(power_timeseries_in_watt: pd.Series, timeresolution: int) -> float:
        """Computes the energy in kWh from a power timeseries in W."""
        if power_timeseries_in_watt.empty:
            return 0.0

        energy_in_kilowatt_hour = float(power_timeseries_in_watt.sum() * timeresolution / 3.6e6)
        return energy_in_kilowatt_hour

    @staticmethod
    def calc_mean_max_min_value(list_or_pandas_series: Union[List, pd.Series]) -> Tuple[float, float, float]:
        """Calc mean, max and min values from List or pd.Series with numpy."""

        mean_value = float(np.mean(list_or_pandas_series))
        max_value = float(np.max(list_or_pandas_series))
        min_value = float(np.min(list_or_pandas_series))

        return mean_value, max_value, min_value



""" Defines the component class and helpers.

The component class is the base class for all other components.
"""
# clean

from __future__ import annotations
import os
import dataclasses as dc
import typing
from dataclasses import dataclass
from typing import Any, Dict, List, Optional
import json
import pandas as pd
from dataclass_wizard import JSONWizard

from hisim import loadtypes as lt
from hisim import log
from hisim.sim_repository import SimRepository
from hisim.simulationparameters import SimulationParameters
from hisim.postprocessing.kpi_computation.kpi_structure import KpiEntry, KpiTagEnumClass
# Package


@dataclass
class ConfigBase(JSONWizard):

    """Base class for all configurations."""

    building_name: str
    name: str

    def __init__(self, name: str, building_name: str = "BUI1"):
        """Initializes."""
        self.building_name = building_name
        self.name = name

    @classmethod
    def get_main_classname(cls):
        """Returns the fully qualified class name for the class that is getting configured. Used for Json."""
        raise NotImplementedError("Missing a definition of the ")

    @classmethod
    def get_config_classname(cls):
        """Gets the class name. Helper function for default connections."""
        return cls.__module__ + "." + cls.__name__

    def get_string_dict(self) -> List[str]:
        """Turns the config into a str list for the report."""
        my_dict = self.to_dict()
        my_list = []
        if len(my_dict) > 0:
            for entry in my_dict.items():
                first_entry = entry[0].rsplit("_")
                first_entry = " ".join(first_entry)
                first_entry = first_entry.capitalize()
                my_list.append(first_entry + ": " + str(entry[1]))
        return my_list


@dataclass
class ComponentConnection:

    """Used in the component class for defining a connection."""

    target_input_name: str
    source_class_name: str
    source_output_name: str
    source_instance_name: Optional[str] = None


class ComponentOutput:  # noqa: too-few-public-methods

    """Used in the component class for defining an output."""

    def __init__(
        self,
        object_name: str,
        field_name: str,
        load_type: lt.LoadTypes,
        unit: lt.Units,
        postprocessing_flag: Optional[List[Any]] = None,
        sankey_flow_direction: Optional[bool] = None,
        output_description: Optional[str] = None,
    ):
        """Defines a component output."""
        self.full_name: str = object_name + " # " + field_name
        self.component_name: str = object_name
        self.field_name: str = field_name
        self.display_name: str = field_name
        self.load_type: lt.LoadTypes = load_type
        self.unit: lt.Units = unit
        self.global_index: int = -1
        self.postprocessing_flag: Optional[List[Any]] = postprocessing_flag
        self.sankey_flow_direction: Optional[bool] = sankey_flow_direction
        self.output_description: Optional[str] = output_description

    def get_pretty_name(self) -> str:
        """Gets a pretty name for a component output."""
        return self.component_name + " - " + self.display_name + " [" + self.load_type + " - " + self.unit + "]"


class ComponentInput:  # noqa: too-few-public-methods

    """Used in the component class for defining an input."""

    def __init__(
        self,
        object_name: str,
        field_name: str,
        load_type: lt.LoadTypes,
        unit: lt.Units,
        mandatory: bool,
    ):
        """Initializes a component input."""
        self.fullname: str = object_name + " # " + field_name
        self.component_name: str = object_name
        self.field_name: str = field_name
        self.loadtype: lt.LoadTypes = load_type
        self.unit: lt.Units = unit
        self.global_index: int = -1
        self.src_object_name: Optional[str] = None
        self.src_field_name: Optional[str] = None
        self.source_output: Optional[ComponentOutput] = None
        self.is_mandatory = mandatory


class SingleTimeStepValues:

    """Contains the values for a single time step."""

    def __init__(self, number_of_values: int):
        """Initializes a new single time step values class."""
        self.values = [0.0] * number_of_values

    def copy_values_from_other(self, other):
        """Copy all values from a single time step values."""
        self.values = other.values[:]

    def clone(self):
        """Makes a copy of the current object."""
        newstsv = SingleTimeStepValues(len(self.values))
        newstsv.values = self.values[:]
        return newstsv

    def get_input_value(self, component_input: ComponentInput) -> float:
        """Gets a value for an input from the single time step values."""
        if component_input.source_output is None:
            return 0
        return self.values[component_input.source_output.global_index]

    def set_output_value(self, output: ComponentOutput, value: float) -> None:
        """Sets a single output value in the single time step values array."""
        self.values[output.global_index] = value

    def is_close_enough_to_previous(self, previous_values: "SingleTimeStepValues") -> bool:
        """Checks if the values are sufficiently similar to another array."""
        count = len(self.values)
        for i in range(count):
            if abs(previous_values.values[i] - self.values[i]) > 0.0001:
                return False
        return True

    def get_differences_for_error_msg(self, previous_values: Any, outputs: List[ComponentOutput]) -> str:
        """Gets a pretty error message for the differences between two time steps."""
        count = len(self.values)
        error_msg = ""
        for i in range(count):
            if abs(previous_values.values[i] - self.values[i]) > 0.0001:
                error_msg += (
                    outputs[i].get_pretty_name()
                    + " previously: "
                    + f"{previous_values.values[i]:4.2f}"
                    + " currently: "
                    + f"{self.values[i]:4.2f}"
                    + " | "
                )
        return error_msg


@dataclass
class DisplayConfig:

    """Configure how to display this component in postprocessing."""

    pretty_name: str | None = None
    display_in_webtool: bool = False

    @classmethod
    def show(cls, pretty_name):
        """Shortcut for showing in webtool with a specified name."""
        return DisplayConfig(pretty_name, display_in_webtool=True)


class Component:

    """Base class for all components."""

    @classmethod
    def get_classname(cls):
        """Gets the class name. Helper function for default connections."""
        return cls.__name__

    @classmethod
    def get_full_classname(cls):
        """Gets the class name. Helper function for default connections."""
        return cls.__module__ + "." + cls.__name__

    def __init__(
        self,
        name: str,
        my_simulation_parameters: SimulationParameters,
        my_config: ConfigBase,
        my_display_config: DisplayConfig,
    ) -> None:
        """Initializes the component class."""
        self.component_name: str = name
        self.inputs: List[ComponentInput] = []
        self.outputs: List[ComponentOutput] = []
        self.outputs_initialized: bool = False
        self.inputs_initialized: bool = False
        self.my_simulation_parameters: SimulationParameters = my_simulation_parameters
        if my_simulation_parameters is None:
            raise ValueError("My Simulation parameters was None.")
        self.simulation_repository: SimRepository
        # self.singleton_simulation_repository: SingletonSimRepository
        self.default_connections: Dict[str, List[ComponentConnection]] = {}
        if isinstance(my_config, ConfigBase):
            self.config = my_config
        else:
            raise ValueError(
                "The argument my_config is not a ConfigBase object.",
                "Please check your components' configuration classes and inherit from ConfigBase class according to hisim/components/example_component.py.",
            )
        self.my_display_config: DisplayConfig = my_display_config

    def get_component_name(self,):
        """Create component name."""
        if self.my_simulation_parameters.multiple_buildings:
            name = str(self.config.building_name) + "_" + str(self.config.name)
        else:
            name = str(self.config.name)
        return name

    def add_default_connections(self, connections: List[ComponentConnection]) -> None:
        """Adds a default connection list definition."""

        component_name = connections[0].source_class_name
        for connection in connections:
            if connection.source_class_name != component_name:
                raise ValueError("Trying to add connections to different components in one go.")
        self.default_connections[component_name] = connections
        log.trace(
            "added default connections for connections from : " + component_name + "\n" + str(self.default_connections)
        )

    def i_prepare_simulation(self) -> None:
        """Gets called before the simulation to prepare the calculation."""
        raise NotImplementedError(
            "Simulation preparation is missing for " + self.component_name + " (" + self.get_full_classname() + ")"
        )

    def set_sim_repo(self, simulation_repository: SimRepository) -> None:
        """Sets the SimRepository."""
        if simulation_repository is None:
            raise ValueError("simulation repository was none")
        self.simulation_repository = simulation_repository

    def add_input(
        self,
        object_name: str,
        field_name: str,
        load_type: lt.LoadTypes,
        unit: lt.Units,
        mandatory: bool,
    ) -> ComponentInput:
        """Adds an input definition."""
        myinput = ComponentInput(object_name, field_name, load_type, unit, mandatory)
        self.inputs.append(myinput)
        return myinput

    def add_output(
        self,
        object_name: str,
        field_name: str,
        load_type: lt.LoadTypes,
        unit: lt.Units,
        postprocessing_flag: Optional[List[Any]] = None,
        sankey_flow_direction: Optional[bool] = None,
        output_description: Optional[str] = None,
    ) -> ComponentOutput:
        """Adds an output definition."""
        if output_description is None:
            raise ValueError("Missing an output description for " + object_name + " - " + field_name)
        log.debug("adding output: " + field_name + " to component " + object_name)
        outp = ComponentOutput(
            object_name,
            field_name,
            load_type,
            unit,
            postprocessing_flag,
            sankey_flow_direction,
            output_description,
        )
        self.outputs.append(outp)
        return outp

    def connect_input(self, input_fieldname: str, src_object_name: str, src_field_name: str) -> None:
        """Connecting an input to an output."""
        if len(self.inputs) == 0:
            raise ValueError("The component " + self.component_name + " has no inputs.")

        component_input: ComponentInput
        input_to_set = None
        for component_input in self.inputs:
            if component_input.field_name == input_fieldname:
                if input_to_set is not None:
                    raise ValueError(
                        "The input "
                        + input_fieldname
                        + " of the component "
                        + self.component_name
                        + " was already set."
                    )
                input_to_set = component_input
        if input_to_set is None:
            raise ValueError("The component " + self.component_name + " has no input with the name " + input_fieldname)
        input_to_set.src_object_name = src_object_name
        input_to_set.src_field_name = src_field_name

        # write input and output connection to json file
        file_name = os.path.join(self.my_simulation_parameters.result_directory, "component_connections.json")

        dict_with_connection_information = {
            "From": {"Component": input_to_set.src_object_name, "Field": input_to_set.src_field_name},
            "To": {"Component": input_to_set.component_name, "Field": input_to_set.field_name},
        }

        if os.path.exists(file_name):
            with open(file_name, mode="r+", encoding="utf-8") as file:
                file.seek(os.stat(file_name).st_size - 1)
                file.write(f",{json.dumps(dict_with_connection_information)}]")
        else:
            with open(file_name, "a", encoding="utf-8") as file:
                json.dump([dict_with_connection_information], file)

    def connect_dynamic_input(self, input_fieldname: str, src_object: ComponentOutput) -> None:
        """For connecting an input to a dynamic output."""
        src_object_name = src_object.component_name
        src_field_name = src_object.field_name
        self.connect_input(
            input_fieldname=input_fieldname,
            src_object_name=src_object_name,
            src_field_name=src_field_name,
        )

    # added variable input length and loop to be able to set default connections in one line in system_setups
    def connect_only_predefined_connections(self, *source_components):
        """Wrapper for default connections and connect with connections list."""
        for source_component in source_components:
            connections = self.get_default_connections(source_component)
            self.connect_with_connections_list(connections)

    def connect_with_connections_list(self, connections: List[ComponentConnection]) -> None:
        """Connect all inputs based on a connections list."""
        for connection in connections:
            src_name: str = typing.cast(str, connection.source_instance_name)
            self.connect_input(connection.target_input_name, src_name, connection.source_output_name)

    def get_default_connections(self, source_component: Component) -> List[ComponentConnection]:
        """Gets the default connections for this component."""
        source_classname: str = source_component.get_classname()
        target_classname: str = self.get_classname()

        if source_classname not in self.default_connections:
            raise ValueError(
                "No default connections for "
                + source_classname
                + " in the connections for "
                + target_classname
                + ". content:\n"
                + str(self.default_connections)
            )
        connections = self.default_connections[source_classname]
        new_connections: List[ComponentConnection] = []
        for connection in connections:
            connection_copy = dc.replace(connection)
            connection_copy.source_instance_name = source_component.component_name
            new_connections.append(connection_copy)
        return new_connections

    def get_input_definitions(self) -> List[ComponentInput]:
        """Gets the input definitions."""
        return self.inputs

    def get_outputs(self) -> List[ComponentOutput]:
        """Delivers a list of outputs."""
        if len(self.outputs) == 0:
            raise ValueError("Error: Component " + self.component_name + " has no outputs defined")
        return self.outputs

    def get_cost_opex(
        self,
        all_outputs: List,
        postprocessing_results: pd.DataFrame,
    ) -> OpexCostDataClass:
        # pylint: disable=unused-argument
        """Calculates operational cost, operational co2 footprint and consumption in kWh (for Diesel in l) during simulation time frame."""
        raise NotImplementedError(f"{self.component_name} has no opex costs implemented.")

    @staticmethod
    def get_cost_capex(config: ConfigBase, simulation_parameters: SimulationParameters) -> CapexCostDataClass:
        # pylint: disable=unused-argument
        """Calculates lifetime, total capital expenditure cost and total co2 footprint of production of device."""
        raise NotImplementedError(f"{config.get_main_classname()} has no capex costs implemented.")

    def get_component_kpi_entries(
        self,
        all_outputs: List,  # pylint: disable=unused-argument
        postprocessing_results: pd.DataFrame,  # pylint: disable=unused-argument
    ) -> List[KpiEntry]:
        """Calculates KPIs for the respective component and return all KPI entries as list."""
        # if the method is not implemented in the component return an empty list
        raise NotImplementedError(f"{self.component_name} has no kpis implemented.")

    def calc_maintenance_cost(self) -> float:
        """Calc maintenance_cost per simulated period as share of capex of component."""
        seconds_per_year = 365 * 24 * 60 * 60
        investment = self.get_cost_capex(config=self.config, simulation_parameters=self.my_simulation_parameters).capex_investment_cost_in_euro

        # add maintenance costs per simulated period
        maintenance_cost_per_simulated_period_in_euro: float = (
            self.config.maintenance_cost_as_percentage_of_investment
            * investment
            * (self.my_simulation_parameters.duration.total_seconds() / seconds_per_year)
        )
        return maintenance_cost_per_simulated_period_in_euro

    def i_save_state(self) -> None:
        """Abstract. Gets called at the beginning of a timestep to save the state."""
        raise NotImplementedError()

    def i_restore_state(self) -> None:
        """Abstract. Restores the state of the component. Can be called many times while iterating."""
        raise NotImplementedError()

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Performs the actual calculation."""
        raise NotImplementedError()

    def write_to_report(self) -> Any:
        """Abstract function for writing the report entry for this component."""
        raise NotImplementedError("In " + self.component_name)

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Abstract. Gets called after the iterations are finished at each time step for potential debugging purposes."""
        pass  # noqa


@dataclass
class OpexCostDataClass:

    """Return element of type OpexCostDataClass in function get_opex_cost from Component."""

    opex_energy_cost_in_euro: float
    opex_maintenance_cost_in_euro: float
    co2_footprint_in_kg: float
    consumption_in_kwh: float
    loadtype: lt.LoadTypes
    kpi_tag: Optional[KpiTagEnumClass]

    @classmethod
    def get_default_opex_cost_data_class(cls) -> OpexCostDataClass:
        """Return the Default for all Components without Opex Costs."""
        return OpexCostDataClass(
            opex_energy_cost_in_euro=0,
            opex_maintenance_cost_in_euro=0,
            co2_footprint_in_kg=0,
            consumption_in_kwh=0,
            loadtype=lt.LoadTypes.ANY,
            kpi_tag=None
        )


@dataclass
class CapexCostDataClass:

    """Return element of type CapexCostDataClass in function get_capex_cost from Component."""

    capex_investment_cost_in_euro: float
    device_co2_footprint_in_kg: float
    lifetime_in_years: float
    capex_investment_cost_for_simulated_period_in_euro: float
    device_co2_footprint_for_simulated_period_in_kg: float
    kpi_tag: Optional[KpiTagEnumClass] = None

    @classmethod
    def get_default_capex_cost_data_class(cls) -> CapexCostDataClass:
        """Return the Default for all Components without Capex Costs."""
        return CapexCostDataClass(
            capex_investment_cost_in_euro=0,
            device_co2_footprint_in_kg=0,
            lifetime_in_years=1,
            capex_investment_cost_for_simulated_period_in_euro=0,
            device_co2_footprint_for_simulated_period_in_kg=0,
            kpi_tag=None
        )
